from unittest import TestCase

import numpy as np
import pandas as pd

from numpy.testing import assert_allclose
from astropy.coordinates import SkyCoord
from astropy import units as u
from astropy.time import Time
from astropy.io import ascii
from km3net_testdata import data_path

import km3astro.coord as kc
import km3astro.frame as kf

from km3astro.random import random_date


class TestCoord(TestCase):
    def setUp(self):
        self.n_evts = 100
        self.n_evts_funny = 1e2

    def test_neutrino_flip_degree(self):
        phi = np.array([97.07, 23.46, 97.07, 192.5, 333.33])
        theta = np.array([135.0, 11.97, 22.97, 33.97, 85.23])
        azi_exp = np.array([277.07, 203.46, 277.07, 12.5, 153.33])
        zen_exp = np.array([45.0, 168.03, 157.03, 146.03, 94.77])
        azi, zen = kc.neutrino_to_source_direction(phi, theta, radian=False)
        assert_allclose(azi, azi_exp)
        assert_allclose(zen, zen_exp)

    def test_neutrino_flip_radian(self):
        phi = np.array([97.07, 23.46, 97.07, 192.5, 333.33]) * np.pi / 180
        theta = np.array([135.0, 11.97, 22.97, 33.97, 85.23]) * np.pi / 180
        azi_exp = np.array([277.07, 203.46, 277.07, 12.5, 153.33]) * np.pi / 180
        zen_exp = np.array([45.0, 168.03, 157.03, 146.03, 94.77]) * np.pi / 180
        azi, zen = kc.neutrino_to_source_direction(phi, theta, radian=True)

        assert_allclose(azi, azi_exp)
        assert_allclose(zen, zen_exp)


class TestCoordRandom(TestCase):
    def test_sun(self):
        date = random_date(n=100)
        sun = kc.sun_local(date, loc="orca")


class TestConvergenceAngle(TestCase):
    def test_convergence_angle(self):
        ca = kf.convergence_angle(1.5, 1.3)
        self.assertAlmostEqual(-0.00897440033130838, ca)


class TestUTMStuff(TestCase):
    def test_utm_zone(self):
        assert 38 == kf.utm_zone(np.pi / 180 * 42.8871)

    def test_longitude_of_central_meridian(self):
        self.assertAlmostEqual(0.785398163397448, kf.longitude_of_central_meridian(38))


class TestAntaresBenchmark(TestCase):
    def setUp(self):
        self.tol = 0.01 * u.deg
        self.gal_tol = 0.02 * u.deg

    def test_antares_objects(self):
        # FIXME
        antares_objects_data = ascii.read(
            data_path("astro/antares_astro_objects_benchmark.csv")
        )
        for obj in antares_objects_data:
            time = Time(" ".join([obj["date"], obj["time"]]))

            theta = np.deg2rad(obj["theta"])
            phi = np.deg2rad(obj["phi"])

            # check azimuth and zenith conversion
            azimuth, zenith = kc.neutrino_to_source_direction(phi, theta)
            self.assertAlmostEqual(azimuth[0], np.deg2rad(obj["azimuth"]))
            self.assertAlmostEqual(zenith[0], np.deg2rad(obj["zenith"]))

            event = kc.local_event(phi, time, theta, location="antares")

            equat = event.fk5
            dec = equat.dec
            ra = equat.ra

            ref = SkyCoord(
                " ".join([obj["RA-J2000"], obj["DEC-J2000"]]),
                unit=(u.hourangle, u.deg),
                frame="fk5",
            )

            # from astropy.coordinates import Angle
            # assert np.abs(Angle(obj["DEC-J2000"] + " hours") - event.fk5.dec) < self.tol
            # assert np.abs(obj["RA-J2000"] * u.deg - event.fk5.ra) < self.tol

            # assert np.abs(dec - ref.fk5.dec) < 0.0001 * u.deg
            # assert np.abs(ra - ref.fk5.ra) < 0.0001 * u.deg

    def test_antares_coordinate_system_benchmarks(self):
        antares_objects_data = ascii.read(
            data_path("astro/antares_coordinate_systems_benchmark.csv")
        )
        for obj in antares_objects_data:
            print(obj)
            time = Time(" ".join([obj["date"], obj["time"]]))

            theta = np.deg2rad(obj["theta"])
            phi = np.deg2rad(obj["phi"])

            # check azimuth and zenith conversion
            azimuth, zenith = kc.neutrino_to_source_direction(phi, theta)
            print("azimuth: ", azimuth, np.rad2deg(azimuth))
            print("zenith: ", zenith, np.rad2deg(zenith))
            self.assertAlmostEqual(azimuth[0], np.deg2rad(obj["azimuth"]))
            self.assertAlmostEqual(zenith[0], np.deg2rad(obj["zenith"]))

            event = kc.local_event(phi, time, theta, location="antares")
            print(event.fk5)
            print(event.galactic)

            # ref = SkyCoord(obj["RA-J2000"], obj["DEC-J2000"], unit=u.deg, frame="fk5")

            assert np.abs(obj["DEC-J2000"] * u.deg - event.fk5.dec) < self.tol
            assert np.abs(obj["RA-J2000"] * u.deg - event.fk5.ra) < self.tol

            print(obj["gal_lat"], event.galactic.b.deg[0])
            assert (
                np.abs(obj["gal_lat"] - event.galactic.b.deg[0]) * u.deg < self.gal_tol
            )
            assert (
                np.abs(obj["gal_lon"] - event.galactic.l.deg[0]) * u.deg < self.gal_tol
            )



import numpy as np
import pandas as pd

import km3astro.coord as kc
import km3astro.frame as kf


def get_az_zenith(SC, detector="antares", unit="deg"):
    """Get the azimuth and zenith for a given sky coordinate

    Parameters
    ----------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    detector : str [default: "antares"]
    The name of the detector, either "antares", "orca" or "arca"
    unit : str
    The unit to be used on the output values, either "deg" or "rad"

    Returns
    -------
    (azimuth, zenith): (float, float)
    Azimuth and zenith in the requested unit
    """

    SC_copy = SC.copy()
    loc = kf.get_location(detector)

    if SC.frame.name != "utm":
        raise ValueError("Wrong Frame: Expected 'utm' but got " + SC.frame.name)

    zenith = SC_copy.zenith.rad
    az = SC_copy.azimuth.rad

    if unit == "deg":
        zenith = SC_copy.zenith.deg
        az = SC_copy.azimuth.deg

    return az, zenith


def get_phi_theta(SC, detector="antares", unit="deg"):
    """Get the phi and theta for a given sky coordinate

    Parameters
    ----------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    detector : str [default: "antares"]
    The name of the detector, either "antares", "orca" or "arca"
    unit : str
    The unit to be used on the output values, either "deg" or "rad"

    Returns
    -------
    (phi, theta): (float, float)
    phi and theta in the requested unit
    """

    SC_copy = SC.copy()
    loc = kf.get_location(detector)

    if SC.frame.name != "particleframe":
        raise ValueError(
            "Wrong Frame: Expected 'particleframe' but got " + SC.frame.name
        )

    phi = SC_copy.phi.rad
    theta = SC_copy.theta.rad

    if unit == "deg":
        phi = SC_copy.phi.deg
        theta = SC_copy.theta.deg

    return phi, theta


def get_alt_az(SC, unit="deg"):
    """Get the altitude and azimuth for a given sky coordinate

    Parameters
    ----------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    unit : str
        The unit to be used on the output values, either "deg" or "rad"

    Returns
    -------
    (altitude, azimuth): (float, float)
    altitude and azimuth in the requested unit
    """

    if SC.frame.name != "altaz":
        raise Exception("Wrong Frame: Expected altAz but got " + SC.frame.name)

    alt = SC.alt
    az = SC.az

    if unit == "deg":
        alt = SC.alt.deg
        az = SC.az.deg

    return alt, az


def get_ra_dec(SC, unit="deg"):
    """Get the right ascension and declination for a given sky coordinate

    Parameters
    ----------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    unit : str
    The unit to be used on the output values, either "deg" or "rad" or "hourangle"

    Returns
    -------
    (ra, dec): (float, float)
    Ra and dec in the requested unit
    """

    if SC.frame.name != "icrs" and SC.frame.name != "fk5":
        raise Exception("Wrong Frame: Expected icrs or fk5 but got " + SC.frame.name)

    ra = SC.ra
    dec = SC.dec

    if unit == "deg":
        ra = SC.ra.deg
        dec = SC.dec.deg

    if unit == "hourangle":
        ra = Angle(ra, unit="hourangle")
        ra = ra.to_string()
        dec = dec

    return ra, dec


def get_l_b(SC, unit="deg"):
    """Get the galactic longitude and galactic latiture for a given sky coordinate

    Parameters
    ----------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    unit : str
    The unit to be used on the output values, either "deg" or "rad"

    Returns
    -------
    (l, b): (float, float)
    l and b in the requested unit
    """

    if SC.frame.name != "galactic":
        raise ValueError("Wrong Frame: Expected galactic but got " + SC.frame.name)

    l = SC.l
    b = SC.b

    if unit == "deg":
        l = SC.l.deg
        b = SC.b.deg

    return l, b


def global_transform(frame_from, frame_to, *args):
    """Global frame transformation function

    Parameters
    ----------
    frame_from : str
        The frame of the input coordinate, either "ParticleFrame" or "UTM" or "equatorial" or "galactic"
    frame_to : str
        The desired frame of the output sky coordinate
    *args : list of the coordinate Parameters
        The sky coordinate parameters. Varies with frame_from and follow "km3astro.coord.build_event" function for reference.

    Returns
    -------
    sky_object: astropy.SkyCoordinate
        The sky coordinate.
    """

    loc = "orca"

    if frame_to == "UTM" or frame_to == "ParticleFrame":
        loc = args[5]

    sky_object = kc.build_event(frame_from, *args)
    sky_object = kc.transform_to(sky_object, frame_to, loc)

    return sky_object


def eq_to_utm(ra, dec, date, time, det):
    """Equatorial to UTM frame transformation function

    Parameters
    ----------
    ra : float
        Right ascension in degree of the sky object
    dec : float
        Declination in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector for the UTM coordinate system, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """

    frame_from = "equatorial"
    frame_to = "UTM"
    SC = global_transform(frame_from, frame_to, date, time, ra, dec, "deg", det)

    return SC


def eq_to_loc(ra, dec, date, time, det):
    """Equatorial to ParticleFrame frame transformation function

    Parameters
    ----------
    ra : float
        Right ascension in degree of the sky object
    dec : float
        Declination in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector for the ParticleFrame coordinate system, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """

    frame_from = "equatorial"
    frame_to = "ParticleFrame"
    SC = global_transform(frame_from, frame_to, date, time, ra, dec, "deg", det)

    return SC


def eq_to_gal(ra, dec, date, time):
    """Equatorial to galactic frame transformation function

    Parameters
    ----------
    ra : float
        Right ascension in degree of the sky object
    dec : float
        Declination in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """

    frame_from = "equatorial"
    frame_to = "galactic"
    SC = global_transform(frame_from, frame_to, date, time, ra, dec, "deg")

    return SC


def loc_to_eq(phi, theta, date, time, det):
    """ParticleFrame to equatorial frame transformation function

    Parameters
    ----------
    phi : float
        Phi in degree of the sky object
    theta : float
        Theta in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector of the sky object, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """

    frame_from = "ParticleFrame"
    frame_to = "equatorial"
    SC = global_transform(frame_from, frame_to, date, time, theta, phi, "deg", det)

    return SC


def loc_to_utm(phi, theta, date, time, det):
    """ParticleFrame to UTM frame transformation function

    Parameters
    ----------
    phi : float
        Phi in degree of the sky object
    theta : float
        Theta in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector of the sky object, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """

    frame_from = "ParticleFrame"
    frame_to = "UTM"
    SC = global_transform(frame_from, frame_to, date, time, theta, phi, "deg", det)

    return SC


def loc_to_gal(phi, theta, date, time, det):
    """ParticleFrame to galactic frame transformation function

    Parameters
    ----------
    phi : float
        Phi in degree of the sky object
    theta : float
        Theta in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector of the sky object, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """

    frame_from = "ParticleFrame"
    frame_to = "galactic"
    SC = global_transform(frame_from, frame_to, date, time, theta, phi, "deg", det)

    return SC


def print_eq_to_utm(ra, dec, date, time, det):
    """Equatorial to UTM frame print function

    Parameters
    ----------
    ra : float
        Right ascension in degree of the sky object
    dec : float
        Declination in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector for the UTM coordinate system, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """
    az, ze = get_az_zenith(eq_to_utm(ra, dec, date, time, det))
    print(str(az) + " " + str(ze))


def print_eq_to_loc(ra, dec, date, time, det):
    """Equatorial to ParticleFrame frame print function

    Parameters
    ----------
    ra : float
        Right ascension in degree of the sky object
    dec : float
        Declination in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector for the ParticleFrame coordinate system, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """
    phi, theta = get_phi_theta(eq_to_loc(ra, dec, date, time, det))
    print(str(phi) + " " + str(theta))


def print_eq_to_gal(ra, dec, date, time):
    """Equatorial to galactic frame print function

    Parameters
    ----------
    ra : float
        Right ascension in degree of the sky object
    dec : float
        Declination in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """
    l, b = get_l_b(eq_to_gal(ra, dec, date, time))
    print(str(l) + " " + str(b))


def print_loc_to_eq(phi, theta, date, time, det):
    """ParticleFrame to equatorial frame print function

    Parameters
    ----------
    phi : float
        Phi in degree of the sky object
    theta : float
        Theta in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector of the sky object, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """
    ra, dec = get_ra_dec(loc_to_eq(phi, theta, date, time, det))
    print(str(ra) + " " + str(dec))


def print_loc_to_utm(phi, theta, date, time, det):
    """ParticleFrame to UTM frame print function

    Parameters
    ----------
    phi : float
        Phi in degree of the sky object
    theta : float
        Theta in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector of the sky object, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """
    az, ze = get_az_zenith(loc_to_utm(phi, theta, date, time, det))
    print(str(az) + " " + str(ze))


def print_loc_to_gal(phi, theta, date, time, det):
    """ParticleFrame to galactic frame print function

    Parameters
    ----------
    phi : float
        Phi in degree of the sky object
    theta : float
        Theta in degree of the sky object
    date : str
        Date of the sky object, format "YYYY-MM-DD"
    time : str
        Time of the sky object, format "HH:MM:SS"
    det : str
    The name of the detector of the sky object, either "antares", "orca" or "arca"

    Returns
    -------
    SC : astropy.SkyCoordinate
        The sky coordinate.
    """
    l, b = get_l_b(loc_to_gal(phi, theta, date, time, det))
    print(str(l) + " " + str(b))


def print_Skycoord(SkyCoord):
    """Print Skycoord object"""
    print(SkyCoord)


def reader_from_file(file):
    """read a csv file into a pandas DataFrame"""
    table = pd.read_csv(file, comment="#")

    return table


def get_frame_name(SC):
    """Get the frame name from a SkyCoord object"""
    return SC.frame.name


def get_obstime(SC):
    """Get the time of observation from a SkyCoord object"""
    return SC.obstime


def split_date_time(dt):
    """Split in date and time a string from the observation time of a SkyCoord object

    Parameters
    ----------
    dt : str
        the date and time string from a SkyCoord object observation time, format "YYYY-MM-DDTHH:MM:SS"

    Returns
    -------
    (date, time): (str, str)
        date and time of the sky object
    """
    date, time = dt.fits.split("T", 1)
    return pd.Series([date, time])


def Skycoord_breaker(data_SC):
    """Break a pandas DataFrame of SkyCoord into a new DataFrame with every parameters (date, time, ra/l/phi/az, dec/b/theta/alt)

    Parameters
    ----------
    data_SC : pandas.DataFrame(SkyCoord)
        pandas DataFrame containing Skycoord object

    Returns
    -------
    data : pandas.DataFrame
        A pandas DataFrame with every parameters of the input SkyCoord object
    """

    sub_sc = pd.DataFrame(data_SC)

    frame = sub_sc.apply(lambda x: get_frame_name(x.SkyCoord_new))
    frame0 = frame[0]

    for i in frame:
        if i != frame0:
            raise ValueError(
                "Non identical frame in list: expect only " + frame0 + " and got " + i
            )

    obstime = sub_sc.apply(lambda x: get_obstime(x.SkyCoord_new))

    data = obstime.apply(lambda x: split_date_time(x))

    data.set_axis(["date", "time"], axis="columns", inplace=True)

    if frame0 == "galactic":
        data["gal_lon"] = sub_sc.apply(lambda x: x.SkyCoord_new.l.deg * 1)
        data["gal_lat"] = sub_sc.apply(lambda x: x.SkyCoord_new.b.deg * 1)
        return data

    elif frame0 == "icrs":
        data["ra"] = sub_sc.apply(lambda x: x.SkyCoord_new.ra.deg * 1)
        data["dec"] = sub_sc.apply(lambda x: x.SkyCoord_new.dec.deg * 1)
        return data

    elif frame0 == "utm":
        data["az"] = sub_sc.apply(lambda x: x.SkyCoord_new.azimuth.deg * 1)
        data["ze"] = sub_sc.apply(lambda x: x.SkyCoord_new.zenith.deg * 1)
        return data

    elif frame0 == "particleframe":
        data["phi"] = sub_sc.apply(lambda x: x.SkyCoord_new.phi.deg * 1)
        data["theta"] = sub_sc.apply(lambda x: x.SkyCoord_new.theta.deg * 1)
        return data

    else:
        raise ValueError("Error: Wrong Skycoord.frame :" + frame0)
        return None


def transform_file(
    file0,
    frame_from,
    frame_to,
    detector_from="orca",
    detector_to="orca",
    name="",
    return_data=False,
):
    """Transform the input file containing coordinate into a new table of those coordinate in a new frame

    Parameters
    ----------
    file0 : str
        The path to the csv file containing the coordinate. Should follow the "km3py/test-data/astro/benchmark" format.
    frame_from : str
        The frame of the coordinate inside file0, either "ParticleFrame" or "UTM" or "equatorial" or "galactic"
    frame_to : str
        The desired frame of the output sky coordinate
    detector_from : str [default = "orca"]
        Detector of the file0 coordinate, either "orca", "arca" or "antares" . Mandatory if using frame_from = "ParticleFrame" or "UTM".
    detector_to : str [default = "orca"]
        Detector to use for the coordinate transformation, either "orca", "arca" or "antares" . Mandatory if using frame_to = "ParticleFrame" or "UTM".
    name : str [default = tempfile]
        path and name of the output file.

    Returns
    -------
    data : csv file
        Csv file containing the pandas dataframe with the transformed coordinate.

    """
    table_read = reader_from_file(file0)

    table_new = kc.transform_to_new_frame(
        table_read, frame_from, frame_to, detector_from, detector_to
    )

    data = [table_new["SkyCoord_new"]]

    data = Skycoord_breaker(data)

    if name == "":
        import tempfile

        path = tempfile.TemporaryDirectory()
        name = "data" + frame_to + ".csv"

    data.to_csv(name, index=False)
    if return_data == True:
        return data


def build_skycoord_list(table, frame, detector="antares"):
    """Transform the input table containing coordinate into a new table of SkyCoord coordinate

    Parameters
    ----------
    table : pandas.DataFrame
        The table from the csv file containing the sky parameter.
    frame : str
        The frame of the coordinate inside the table, either "ParticleFrame" or "UTM" or "equatorial" or "galactic"
    detector : str [default = "antares"]
        Detector of the coordinate inside the table, either "orca", "arca" or "antares" . Mandatory if using frame_from = "ParticleFrame" or "UTM".

    Returns
    -------
    list_evt : pandas.DataFrame(astropy.SkyCoord)
        DataFrame containing astropy.SkyCoord object corresponding to the sky coordinate in the input table.

    """

    if frame == "ParticleFrame":
        list_evt = table.apply(
            lambda x: kc.build_event(
                frame, x.date, x.time, x.theta, x.phi, "deg", detector
            ),
            axis=1,
            result_type="expand",
        )

    if frame == "UTM":
        list_evt = table.apply(
            lambda x: kc.build_event(
                frame, x.date, x.time, x.azimuth, x.zenith, "deg", detector
            ),
            axis=1,
            result_type="expand",
        )

    if frame == "equatorial":
        list_evt = table.apply(
            lambda x: kc.build_event(
                frame, x.date, x.time, x["RA-J2000"], x["DEC-J2000"]
            ),
            axis=1,
            result_type="expand",
        )

    if frame == "galactic":
        list_evt = table.apply(
            lambda x: kc.build_event(frame, x.date, x.time, x.gal_lon, x.gal_lat),
            axis=1,
            result_type="expand",
        )

    if isinstance(list_evt, pd.Series):
        series_ = {"SkyCoord_base": list_evt}
        list_evt = pd.DataFrame(series_)

    list_evt.set_axis(["SkyCoord_base"], axis="columns", inplace=True)

    return list_evt



#!/usr/bin/env python
"""Test whether at some time the sun is below the horizon in ORCA."""

import os.path

import numpy as np
import km3astro
import km3astro.extras

from ..random import random_date


DATADIR = os.path.join(os.path.dirname(km3astro.__file__), "data")


def above_horiz(times):
    pd = km3astro.extras.pandas()

    fname = os.path.join(DATADIR, "orca_sun_isup.h5")
    with pd.HDFStore(fname) as h5:
        o_rise = h5["o_rise"].values
        o_set = h5["o_set"].values
    return [(t >= o_rise) & (t <= o_set) for t in times]


def is_below(times):
    ab = np.array(above_horiz(times))
    risen = np.logical_not(np.sum(ab, axis=1))
    return risen


def random_time_sun_vis(n=1, keep_above=False, **kwargs):
    ngen = n * 3
    dates = random_date(n=ngen, **kwargs)
    mask = is_below(dates)
    if keep_above:
        mask = np.logical_not(mask)
    dates = dates[mask]
    if len(dates) < n:
        dates = random_time_sun_vis(n * 2, keep_above=keep_above, **kwargs)
    return dates[:n]



""" km3astro.testing_tools for tests/test_benchmark.py"""

import numpy as np
import pandas as pd

import km3astro.coord as kc
import km3astro.toolbox as ktb


def test_Skycoord_separation(SC_true, SC_check):
    """Calculate the angle separation between two SkyCoord object

    Parameters
    ----------
    SC_true : astropy.SkyCoord
        Sky coordinate of reference
    SC_check : astropy.SkyCoord
        Sky coordinate to be checked

    Returns
    -------
    sep_deg : float
        The angle between the two Sky coordinate object
    """

    sep = SC_true.separation(SC_check)
    sep_deg = sep.deg
    return sep_deg


def test_benchmark_conversion(table_true, table_check):
    """Calculate the angle separation between two pandas DataFrame of SkyCoord object

    Parameters
    ----------
    table_true : pandas.DataFrame(astropy.SkyCoord)
        Pandas DataFrame of the Sky coordinate of reference
    table_check : pandas.DataFrame(astropy.SkyCoord)
        Pandas DataFrame of the Sky coordinate to be checked

    Returns
    -------
    sep_table : pandas.DataFrame(float)
        DataFrame of the angle between the two table of Sky coordinate object
    """

    data = [table_true["SkyCoord_base"], table_check["SkyCoord_new"]]
    table = pd.concat(data, axis=1)

    sep_table = table.apply(
        lambda x: test_Skycoord_separation(x.SkyCoord_base, x.SkyCoord_new),
        axis=1,
        result_type="expand",
    )

    return sep_table


def test_angle_separation(file0, detector="antares", detector_to="antares"):
    """Test the angle separation of between a benchmark table and km3astro.coord transformation tool.
    for test purpose only.

    Parameters
    ----------
    file0 : str
        path to the benchmark table.
    detector : str [default = "antares"]
        Detector of the benchmark table, either "orca", "arca" or "antares"
    detector_to : str [default = "antares"]
        Detector of the new frame, either "orca", "arca" or "antares"

    """
    table_read = ktb.reader_from_file(file0)

    angle_treshold = 0.02

    if set(["phi", "theta"]).issubset(table_read.columns):

        table_loc_to_utm = kc.transform_to_new_frame(
            table_read, "ParticleFrame", "UTM", detector, detector_to
        )
        table_loc_to_eq = kc.transform_to_new_frame(
            table_read, "ParticleFrame", "equatorial", detector, detector_to
        )
        table_loc_to_gal = kc.transform_to_new_frame(
            table_read, "ParticleFrame", "galactic", detector, detector_to
        )

    if set(["azimuth", "zenith"]).issubset(table_read.columns):

        table_utm_to_loc = kc.transform_to_new_frame(
            table_read, "UTM", "ParticleFrame", detector, detector_to
        )
        table_utm_to_eq = kc.transform_to_new_frame(
            table_read, "UTM", "equatorial", detector, detector_to
        )
        table_utm_to_gal = kc.transform_to_new_frame(
            table_read, "UTM", "galactic", detector, detector_to
        )

    if set(["RA-J2000", "DEC-J2000"]).issubset(table_read.columns):

        table_eq_to_utm = kc.transform_to_new_frame(
            table_read, "equatorial", "UTM", detector, detector_to
        )
        table_eq_to_loc = kc.transform_to_new_frame(
            table_read, "equatorial", "ParticleFrame", detector, detector_to
        )
        table_eq_to_gal = kc.transform_to_new_frame(
            table_read, "equatorial", "galactic", detector, detector_to
        )

    if set(["gal_lon", "gal_lat"]).issubset(table_read.columns):

        table_gal_to_loc = kc.transform_to_new_frame(
            table_read, "galactic", "ParticleFrame", detector, detector_to
        )
        table_gal_to_eq = kc.transform_to_new_frame(
            table_read, "galactic", "equatorial", detector, detector_to
        )
        table_gal_to_utm = kc.transform_to_new_frame(
            table_read, "galactic", "UTM", detector, detector_to
        )

    # testing angle separation
    if set(["phi", "theta"]).issubset(table_read.columns):

        if set(["azimuth", "zenith"]).issubset(table_read.columns):
            sep_utm_to_loc = test_benchmark_conversion(
                table_loc_to_utm, table_utm_to_loc
            )

            mean_ = sep_utm_to_loc.mean()
            min_ = sep_utm_to_loc.min()
            max_ = sep_utm_to_loc.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["RA-J2000", "DEC-J2000"]).issubset(table_read.columns):
            sep_eq_to_loc = test_benchmark_conversion(table_loc_to_eq, table_eq_to_loc)

            mean_ = sep_eq_to_loc.mean()
            min_ = sep_eq_to_loc.min()
            max_ = sep_eq_to_loc.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["gal_lon", "gal_lat"]).issubset(table_read.columns):

            sep_gal_to_loc = test_benchmark_conversion(
                table_loc_to_gal, table_gal_to_loc
            )

            mean_ = sep_gal_to_loc.mean()
            min_ = sep_gal_to_loc.min()
            max_ = sep_gal_to_loc.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

    if set(["azimuth", "zenith"]).issubset(table_read.columns):

        if set(["phi", "theta"]).issubset(table_read.columns):
            sep_loc_to_utm = test_benchmark_conversion(
                table_utm_to_loc, table_loc_to_utm
            )

            mean_ = sep_loc_to_utm.mean()
            min_ = sep_loc_to_utm.min()
            max_ = sep_loc_to_utm.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["RA-J2000", "DEC-J2000"]).issubset(table_read.columns):
            sep_eq_to_utm = test_benchmark_conversion(table_utm_to_eq, table_eq_to_utm)

            mean_ = sep_eq_to_utm.mean()
            min_ = sep_eq_to_utm.min()
            max_ = sep_eq_to_utm.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["gal_lon", "gal_lat"]).issubset(table_read.columns):
            sep_gal_to_utm = test_benchmark_conversion(
                table_utm_to_gal, table_gal_to_utm
            )

            mean_ = sep_gal_to_utm.mean()
            min_ = sep_gal_to_utm.min()
            max_ = sep_gal_to_utm.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

    if set(["RA-J2000", "DEC-J2000"]).issubset(table_read.columns):

        if set(["phi", "theta"]).issubset(table_read.columns):
            sep_loc_to_eq = test_benchmark_conversion(table_eq_to_loc, table_loc_to_eq)

            mean_ = sep_loc_to_eq.mean()
            min_ = sep_loc_to_eq.min()
            max_ = sep_loc_to_eq.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["azimuth", "zenith"]).issubset(table_read.columns):
            sep_utm_to_eq = test_benchmark_conversion(table_eq_to_utm, table_utm_to_eq)

            mean_ = sep_utm_to_eq.mean()
            min_ = sep_utm_to_eq.min()
            max_ = sep_utm_to_eq.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["gal_lon", "gal_lat"]).issubset(table_read.columns):
            sep_gal_to_eq = test_benchmark_conversion(table_eq_to_gal, table_gal_to_eq)

            mean_ = sep_gal_to_eq.mean()
            min_ = sep_gal_to_eq.min()
            max_ = sep_gal_to_eq.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

    if set(["gal_lon", "gal_lat"]).issubset(table_read.columns):

        if set(["phi", "theta"]).issubset(table_read.columns):
            sep_loc_to_gal = test_benchmark_conversion(
                table_gal_to_loc, table_loc_to_gal
            )

            mean_ = sep_loc_to_gal.mean()
            min_ = sep_loc_to_gal.min()
            max_ = sep_loc_to_gal.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["azimuth", "zenith"]).issubset(table_read.columns):
            sep_utm_to_gal = test_benchmark_conversion(
                table_gal_to_utm, table_utm_to_gal
            )

            mean_ = sep_utm_to_gal.mean()
            min_ = sep_utm_to_gal.min()
            max_ = sep_utm_to_gal.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )

        if set(["RA-J2000", "DEC-J2000"]).issubset(table_read.columns):
            sep_eq_to_gal = test_benchmark_conversion(table_gal_to_eq, table_eq_to_gal)

            mean_ = sep_eq_to_gal.mean()
            min_ = sep_eq_to_gal.min()
            max_ = sep_eq_to_gal.max()

            if max_ > angle_treshold:
                raise AssertionError(
                    "Error: Maximum angle separation = "
                    + str(max_)
                    + " > "
                    + str(angle_treshold)
                )



"""Plotting utilities.
"""
from astropy.units import degree

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import healpy as hp
from datetime import timedelta

import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.coordinates import ICRS, Galactic, FK4, FK5  # Low-level frames

from astropy.time import Time

from km3net_testdata import data_path

import km3astro.coord as kc
import km3astro.frame as kf
import km3astro.toolbox as kt
import tempfile
import os


def ra_dec(skycoord):
    """Take (ra, dec) from skycoord in matplotlib-firendly format.

    This wraps the ra because astropy's convention differs from matplotlib.
    """
    ra = skycoord.ra.wrap_at(180 * degree).radian
    dec = skycoord.dec.radian
    return ra, dec


def projection_axes(projection="aitoff", **figargs):
    fig = plt.figure(**figargs)
    ax = fig.add_subplot(111, projection=projection)
    ax.grid(color="lightgrey")
    return fig, ax


def plot_equatorial(
    evts,
    projection="aitoff",
    ax=None,
    marker="o",
    markersize=4,
    alpha=0.8,
    adjust_subplots=True,
    **kwargs,
):
    ra, dec = ra_dec(evts)
    if ax is None:
        _, ax = projection_axes(projection=projection)
    ax.plot(ra, dec, marker, markersize=markersize, alpha=alpha, **kwargs)
    if adjust_subplots:
        plt.subplots_adjust(top=0.95, bottom=0.0)
    return ax


def get_coord_from_skycoord(SC, frame, detector):
    """Return the coordinate of a SkyCoord object for aitoff or mollweide projection

    Parameters
    ----------
    SC : astropy.SkyCoord
        The sky coordinate.
    frame : str
        The frame of the coordinate, either "ParticleFrame" or "UTM" or "altaz" or "equatorial" or "galactic"
    detector : str [default = "antares"]
        Detector of the coordinate, either "orca", "arca" or "antares".

    Returns
    -------
    (phi, theta)/(az, ze)/(alt, az)/(ra, dec)/(l, b) : (float, float)
        The set of coordinate.

    """
    SC_copy = kc.transform_to(SC, frame, detector)

    if frame == "ParticleFrame":
        phi = SC_copy.phi.wrap_at(180 * u.deg).radian

        theta = SC_copy.theta.radian

        if type(theta) == np.float64:
            if theta > np.pi / 2:
                theta = theta - np.pi

        if type(theta) == np.ndarray:
            print(len(theta))
            for idx, t in enumerate(theta):
                if t > np.pi / 2:
                    theta[idx] = t - np.pi

        return phi, theta

    elif frame == "UTM":
        az = SC_copy.azimuth.wrap_at(180 * u.deg).radian
        ze = SC_copy.alt_utm.radian
        return az, ze

    elif frame == "altaz":
        alt = SC_copy.alt.radian
        az = SC_copy.az.wrap_at(180 * u.deg).radian
        return alt, az

    elif frame == "equatorial":
        ra = SC_copy.ra.wrap_at(180 * u.deg).radian
        dec = SC_copy.dec.radian
        return ra, dec

    elif frame == "galactic":
        l = SC_copy.l.wrap_at(180 * u.deg).radian
        b = SC_copy.b.radian
        return l, b

    else:
        raise ValueError("Error: Wrong Frame input frame")
        return None


def plot_SkyCoord(
    SC,
    frame,
    detector,
    projection="aitoff",
    ax=None,
    marker="s",
    markersize=3,
    alpha=0.8,
    adjust_subplots=True,
    **kwargs,
):
    """Plot given SkyCoord object or list"""

    Coord_lon, Coord_lat = get_coord_from_skycoord(SC, frame, detector)

    if ax is None:
        _, ax = projection_axes(projection=projection)
    ax.plot(Coord_lon, Coord_lat, marker, markersize=markersize, alpha=alpha, **kwargs)
    if adjust_subplots:
        plt.subplots_adjust(top=0.95, bottom=0.0)
    return ax


def get_Sky_label(SC, frame, detector, time=False):
    """Return the coordinate of a SkyCoord object as string to use as label for matplotlib.

    Parameters
    ----------
    SC : astropy.SkyCoord
        The sky coordinate.
    frame : str
        The frame of the coordinate, either "ParticleFrame" or "UTM" or "altaz" or "equatorial" or "galactic"
    detector : str [default = "antares"]
        Detector of the coordinate, either "orca", "arca" or "antares".
    time : bool
        Bool to integrate or no the time in the label.

    Returns
    -------
    labels : str
        A string containing the coordinate and time if asked to use as label.

    """

    a, b = get_coord_from_skycoord(SC, frame, detector)
    a = a * 180 / np.pi
    a = float("{:.2f}".format(a))
    b = b * 180 / np.pi
    b = float("{:.2f}".format(b))

    labels = ""

    xlab = ""
    ylab = ""
    timelab = ""

    if time == True:
        timelab = str(SC.obstime)

    if frame == "ParticleFrame":
        xlab = "Phi = " + str(a)
        ylab = " Theta = " + str(b) + " [deg]"

    elif frame == "UTM":
        xlab = "Azimuth = " + str(a)
        ylab = " Zenith = " + str(b) + " [deg]"

    elif frame == "equatorial":
        xlab = "RA = " + str(a)
        ylab = " DEC = " + str(b) + " [deg] (ICRS)"

    elif frame == "galactic":
        xlab = "l = " + str(a)
        ylab = " b = " + str(b) + " [deg]"

    else:
        raise ValueError("Error: Wrong Frame input frame")
        return None

    labels = xlab + ylab + " " + timelab
    return labels


def get_label(plot_frame):
    """Return the Axis label corresponding to the plot frame.

    Parameters
    ----------
    plot_frame : str
        The plot frame.

    Returns
    -------
    (xlab, ylab) : (str,str)
        X and Y axis name to use as label.

    """

    if plot_frame == "ParticleFrame":
        xlab = "Phi / deg"
        ylab = "Theta / deg"
        return xlab, ylab

    elif plot_frame == "UTM":
        xlab = "Azimuth / deg"
        ylab = "Zenith / deg"
        return xlab, ylab

    elif plot_frame == "equatorial":
        xlab = "RA(ICRS) /deg"
        ylab = "DEC(ICRS) / deg"
        return xlab, ylab

    elif plot_frame == "galactic":
        xlab = "Gal_lon /deg"
        ylab = "Gal_lat /deg"
        return xlab, ylab

    else:
        raise ValueError("Error: Wrong Frame input frame")
        return None


def get_horizon(detector, time, frame, alt_cut=5.7):
    """Calculate the Horizon line of view for a given detector, time and frame system.

    Parameters
    ----------
    detector : str
        Detector to use, either "orca", "arca" or "antares".

    time : str or astropy.Time
        The time of observation of the alert.

    frame : str
        The frame to return the calculated horizon, either "ParticleFrame" or "UTM" or "equatorial" or "galactic"
    cut : float[default = 5.7]
        Value in degree of the cut in Altitude for Altaz frame. By default at 5.7 degree corresponding to cos(0.1).

    Returns
    -------
    horizon : np.ndarray(astropy.SkyCoord)
        A list of SkyCoord representing the horizon line of view.

    """

    n = 360
    alts = alt_cut * np.ones(n)
    azs = np.zeros(n)

    for i in range(n):
        azs[i] = i

    obstime = Time(time)
    loc = kf.get_location(detector)

    horizon = SkyCoord(
        alt=alts * u.degree,
        az=azs * u.degree,
        obstime=obstime,
        location=loc,
        frame="altaz",
    )

    horizon = kc.transform_to(horizon, frame, detector)

    return horizon


def get_galactic_plan(detector, time, frame):
    """Calculate the Galactic plan for a given detector, time and frame system.

    Parameters
    ----------
    detector : str
        Detector to use, either "orca", "arca" or "antares".

    time : str or astropy.Time
        The time of observation of the alert.

    frame : str
        The frame to return the calculated horizon, either "ParticleFrame" or "UTM" or "equatorial" or "galactic"

    Returns
    -------
    gal_plan : np.ndarray(astropy.SkyCoord)
        A list of SkyCoord representing the galactic plan.

    """
    n = 360
    l = np.zeros(n)
    b = np.zeros(n)

    for i in range(n):
        l[i] = i

    obstime = Time(time)
    loc = kf.get_location(detector)

    gal_plan = SkyCoord(
        l=l * u.degree, b=b * u.degree, obstime=obstime, location=loc, frame="galactic"
    )

    gal_plan = kc.transform_to(gal_plan, frame, detector)

    return gal_plan


def get_obstime(table_Sky, it=0):
    """return the observation time for a table of SkyCoord at given index."""
    return table_Sky.iloc[it]["SkyCoord_base"].obstime


def get_alert_color(alert_type):
    """Return the color for a specific alert_type
    Based on color list tab:Palette 10
    """

    Alert_color_dict = {
        "GRB": "tab:blue",
        "GW": "tab:orange",
        "Neutrino": "tab:green",
        "NuEM": "tab:red",
        "SK_SN": "tab:purple",
        "SNEWS": "tab:brown",
        "Transient": "tab:pink",
    }

    if alert_type in Alert_color_dict.keys():
        return Alert_color_dict[alert_type]
    else:
        return "c"


def plot_pd_skycoord(table_skycoord, plot_frame, detector_to, projection, ax):
    """plot a pandas DataFrame of SkyCoord."""

    if "Alert_type" in table_skycoord.columns:
        table_skycoord.apply(
            lambda x: plot_SkyCoord(
                x.SkyCoord_base,
                frame=plot_frame,
                detector=detector_to,
                projection=projection,
                markersize=10,
                marker=".",
                color=get_alert_color(x.Alert_type),
                ax=ax,
                label=get_Sky_label(x.SkyCoord_base, plot_frame, detector_to),
            ),
            axis=1,
        )

    else:
        table_skycoord.apply(
            lambda x: plot_SkyCoord(
                x.SkyCoord_base,
                frame=plot_frame,
                detector=detector_to,
                projection=projection,
                markersize=10,
                marker=".",
                color="royalblue",
                ax=ax,
                label=get_Sky_label(x.SkyCoord_base, plot_frame, detector_to),
            ),
            axis=1,
        )


def plot_file(file0, ax, projection, frame, detector, plot_frame, detector_to):
    """plot a csv file containing alerts."""
    table_read = pd.read_csv(file0, comment="#")
    table_skycoord = kt.build_skycoord_list(table_read, frame, detector)
    plot_pd_skycoord(table_skycoord, plot_frame, detector_to, projection, ax)


def calculate_visibility_map(detector="orca", frame="equatorial", alt_cut=5.7):
    """Calculate the visibility map of a detector in a frame for a given cut on altitude.
       Warning: Time of Calcultation can be really long.

    Parameters
    ----------

    detector : str [default=orca]
        The detector to use for calculation, either "orca", "arca" or "antares.
    frame : str [default="equatorial"]
        The frame to use for the visibility map, either "equatorial" or "galactic"
    alt_cut : float [default=5.7]
        The cut in altitude of the detector line of view. default is 5.7 degree corresponding to cos(0.1)

    Returns
    -------

    visi_map : csv
        Write a file containing the visibility map table.
        Already included visibility_map: antares, orca, arca in equatorial and galactic

    """

    Ntime = 24
    day = 24

    visi_map = np.zeros((360, 180))
    counter = 0
    print(len(visi_map))
    print(len(visi_map[0]))
    for i in range(len(visi_map)):

        for j in range(len(visi_map[i])):
            counter += 1
            print(str(counter) + "/" + str(len(visi_map) * len(visi_map[0])))
            prob = 0
            for k in range(Ntime):

                sec = 24 * 60 * 60 / Ntime
                dtime = timedelta(seconds=sec * k)
                date = "2022-01-01"
                time = str(dtime)

                SC = kt.global_transform(
                    frame, "altaz", date, time, i, (j - 90), "deg", detector
                )
                if SC.alt.deg < alt_cut:
                    prob += 1
            print("Prob = " + str(prob))
            prob = prob / Ntime
            visi_map[i][j] = prob

    path = "skymap_plot/"
    name = "visibility_map_" + frame + "_" + detector + ".csv"

    pd.DataFrame(visi_map).to_csv(path + name, index=None, header=None)


def read_visibility_map(visi_map):
    """Read a given visibility map"""

    data = pd.read_csv(visi_map, sep=",", header=None)
    return data


def get_visi_map_path(frame, detector):
    """Return the path to a visibility map for a given frame and detector"""
    path = (
        os.path.dirname(os.path.abspath(__file__))
        + "/data/visibility_map_"
        + frame
        + "_"
        + detector
        + ".csv"
    )
    return path


def plot_visibility(ax, frame="equatorial", detector="antares", plot_colorscale=False):
    """Plot on the background the visibility map corresponding to the frame and detector"""
    visi_map = get_visi_map_path(frame, detector)

    visi_data = read_visibility_map(visi_map).to_numpy()

    x = np.arange(-90, 90, 1)
    y = np.arange(-180, 180, 1)
    x = x * (2 * np.pi / 360)
    y = y * (2 * np.pi / 360)

    pc = ax.pcolormesh(y, x, visi_data.T, alpha=1, cmap="Greys_r", shading="gouraud")
    if plot_colorscale:
        plt.colorbar(pc, shrink=0.6)


def skymap_hpx(file0, save=False, path="", name=""):
    # to be updated for GW healpix url
    """Method to plot a skymap from an hpx url

    Parameters
    ----------
    file0 : str
        The path to the healpix file.
    save : bool
        To save or not the fig
    path : str
        Path to the writing location.
    name : str
        Path and name of where to write the skymap png file.

    Returns:
    Fig : file.png
       A png file of the skymap.
    """

    projection = "mollweide"
    fig, ax = projection_axes(projection=projection, figsize=[40 / 2.54, 30 / 2.54])

    gw_map = hp.read_map(file0)
    hp.mollview(
        map = gw_map,
        fig = fig
    )
    hp.graticule()

    if save:
        if path != "":
            if name == "":
                name = os.path.join(path.name, f"skymap_hpx_test.png")

            else:
                name = os.path.join(path.name, f"{name}.png")

        else:
            path = tempfile.TemporaryDirectory()

            if name == "":
                name = "hpx_skymap_maker_test.png"

            plt.savefig(name)

    return fig


def skymap_list(
    file0="",
    dataframe=pd.DataFrame(),
    frame="equatorial",
    detector="antares",
    plot_frame="equatorial",
    detector_to="antares",
    title="",
    save=False,
    path="",
    name="",
):
    """Method to plot a skymap from a list of alert in a csv file.

    Parameters
    ----------
    file0 : str
        The path to the csv file containing the list of alert.
    dataframe = pd.DataFrame()
        The dataframe containing the list of alert
    frame :str [default = "equatorial"]
        The frame of alerts in file0
    detector : str [default = "antares"]
        The detector of the alerts, either "orca", "arca" or "antares"
    plot_frame : str [default = "equatorial"]
        The frame of the skymap, either "equatorial" or "galactic"
    detector_to : str [default = "antares"]
        The detector to use for frame transformation, either "orca", "arca" or "antares"
    title : str [default = ""]
        Title of the figure
    save : bool [default = False]
        To save or not the figure
    path : str
        Path to the writing location
    name : str
        Name the skymap png file.

    Returns:
    Fig : file.png
       A png file of the skymap.
    """

    projection = "aitoff"
    fig, ax = projection_axes(projection=projection, figsize=[40 / 2.54, 30 / 2.54])

    if file0 != "":
        table_read = pd.read_csv(file0, comment="#")
        table_skycoord = kt.build_skycoord_list(table_read, frame, detector)

    elif dataframe.empty == False:
        table_skycoord = kt.build_skycoord_list(dataframe, frame, detector)
        if "Alert_type" in dataframe.columns:
            extracted_column = dataframe["Alert_type"]
            table_skycoord = table_skycoord.join(extracted_column)

    else:
        file0 = data_path("astro/antares_coordinate_systems_benchmark.csv")
        table_skycoord = kt.build_skycoord_list(table_read, frame, detector)

    plot_pd_skycoord(table_skycoord, plot_frame, detector_to, projection, ax)
    plot_visibility(ax=ax, frame=plot_frame, detector=detector_to)

    if title == "":
        title = detector + " Alert List " + plot_frame + " frame skymap"
    ax.set_title(title, fontsize=20, y=1.1)
    xlabel, ylabel = get_label(plot_frame)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    if "Alert_type" in table_skycoord.columns:

        h = []
        check_list = []

        for it in table_skycoord["Alert_type"]:

            if it not in check_list:
                check_list.append(it)

                alert = mlines.Line2D(
                    [],
                    [],
                    color=get_alert_color(it),
                    marker=".",
                    markersize=10,
                    label=it,
                )

                h.append(alert)

        plt.legend(handles=h, bbox_to_anchor=(1.05, 1.05))

    if save:
        if path != "":
            if name == "":
                name = os.path.join(
                    path.name, f"skymap_list_{detector}_test_{plot_frame}.png"
                )

            else:
                name = os.path.join(path.name, f"{name}.png")

        else:
            path = tempfile.TemporaryDirectory()

            if name == "":
                name = "skymap_list_" + detector + "_test_" + plot_frame + ".png"

        plt.savefig(name)

    return fig


def skymap_alert(
    file0="",
    ra=1000,
    dec=1000,
    obstime="",
    frame="equatorial",
    detector="antares",
    plot_frame="equatorial",
    detector_to="antares",
    alt_cut=5.7,
    title="",
    save=False,
    path="",
    name="",
):
    """Method to plot a skymap from an alert in a csv file or by giving RA, DEC and obstime.

    Parameters
    ----------
    file0 : str [default = ""]
        The path to the csv file containing the alert.
    ra, dec : (float,float)
        The ra and dec coordinate of the alert.
    obstime : str
        The observation time of the alert. Format is "YYYY-MM-DDTHH:MM:SS"
    frame :str [default = "equatorial"]
        The frame of the alert.
    detector : str [default = "antares"]
        The detector of the alert.
    plot_frame : str [default = "equatorial"]
        The frame of the skymap, either "ParticleFrame", "UTM", "equatorial" or "galactic"
    detector_to :str [default = "antares"]
        The detector to use for frame transformation, either "orca", "arca" or "antares"
    alt_cut : float [default = 5.7]
        The altitude cut on detector visibility for horizon calculation. default is alt_cut = 5.7 degree which correspond to cos(0.1).
    title : str [default = ""]
        Title of the figure
    save : bool [default = False]
        To save or not the figure
    path : str
        Path to the writing location.
    name : str [default = ""]
        Name of the skymap png file.

    Returns:
    Fig : file.png
       A png file of the skymap.
    """

    use_coord = False
    use_file = False

    if file0 != "":
        use_file = True

    if ra != 1000 or dec != 1000 or obstime != "":
        use_coord = True

    if use_coord == False and use_file == False:
        file0 = data_path("astro/antares_moon_sun_position_benchmark.csv")
        ra = 80
        dec = 15
        obstime = "2022-06-15T03:03:03"
        use_coord = True
        use_file = True

    projection = "aitoff"
    fig, ax = projection_axes(projection=projection, figsize=[40 / 2.54, 30 / 2.54])

    if use_file == True:

        table_read = pd.read_csv(file0, comment="#")
        table_skycoord = kt.build_skycoord_list(table_read, frame, detector)
        obstime = get_obstime(table_skycoord)
        plot_file(file0, ax, projection, frame, detector, plot_frame, detector_to)

    if use_coord == True:

        alert = SkyCoord(
            ra=ra * u.degree, dec=dec * u.degree, obstime=obstime, frame="icrs"
        )

        plot_SkyCoord(
            alert,
            frame=plot_frame,
            detector=detector_to,
            projection=projection,
            ax=ax,
            marker=".",
            markersize=10,
            linewidth=0,
            color="royalblue",
            label=get_Sky_label(alert, plot_frame, detector_to, time=False),
        )

    horizon = get_horizon(
        detector=detector_to, time=obstime, frame=plot_frame, alt_cut=alt_cut
    )

    plot_SkyCoord(
        horizon,
        frame=plot_frame,
        detector=detector_to,
        projection=projection,
        ax=ax,
        marker=".",
        markersize=10,
        linewidth=0,
        color="darkgreen",
    )

    gal_plan = get_galactic_plan(detector=detector_to, time=obstime, frame=plot_frame)
    plot_SkyCoord(
        gal_plan,
        frame=plot_frame,
        detector=detector_to,
        projection=projection,
        ax=ax,
        marker=".",
        markersize=10,
        linewidth=0,
        color="darkred",
    )

    date, time = str(obstime).split("T", 1)

    if title == "":
        title = detector + " Alert " + plot_frame + " frame skymap " + date + " " + time
    ax.set_title(title, fontsize=20, y=1.1)
    xlabel, ylabel = get_label(plot_frame)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

    horizon_line = mlines.Line2D(
        [], [], color="lime", marker=".", markersize=1, label="Horizon"
    )
    gal_line = mlines.Line2D(
        [], [], color="red", marker=".", markersize=1, label="Galactic plan"
    )

    h, l = ax.get_legend_handles_labels()
    h.append(horizon_line)
    h.append(gal_line)

    plt.legend(handles=h, bbox_to_anchor=(1.05, 1.05))

    if save:
        if path != "":
            if name == "":
                name = os.path.join(
                    path.name, f"skymap_alert_{detector}_test_{plot_frame}.png"
                )
            else:
                name = os.path.join(path.name, f"{name}.png")

        else:
            path = tempfile.TemporaryDirectory()
            if name == "":
                name = "skymap_alert_" + detector + "_test_" + plot_frame + ".png"

        plt.savefig(name)

    return fig



"""
========================
Multiple exemple of how to use km3astro.toolbox
========================

"""

# Author: Tedjditi Hichem <htedjditi@km3net.de>

import km3astro.toolbox as ktb
from km3net_testdata import data_path


def main():

    ##########################################################
    # Alert Coordinates
    # --------------------
    # Let's define one event.

    det = "antares"
    date = "2007-10-04"
    time = "03:03:03"
    phi = 97.07
    theta = 135.0
    az = 277.07
    ze = 45.00
    ra = 70.613
    dec = -1.852
    l = 198.7
    b = -29.298

    ##########################################################
    # Let's do some transformation with toolbox.function

    print("Starting test of toolbox")
    print()
    print("TESTING PRINT EQ TO XXX")

    ktb.print_eq_to_utm(ra, dec, date, time, det)
    ktb.print_eq_to_loc(ra, dec, date, time, det)
    ktb.print_eq_to_gal(ra, dec, date, time)
    print()
    print("TESTING PRINT LOC TO XXX")

    ktb.print_loc_to_eq(phi, theta, date, time, det)
    ktb.print_loc_to_utm(phi, theta, date, time, det)
    ktb.print_loc_to_gal(phi, theta, date, time, det)
    print()
    print("TESTING TRANSFORM FILE")

    ##########################################################
    # Let's transform a file with a list of alert
    # --------------------
    ##########################################################
    # Let's take one list of event.

    file0 = data_path("astro/antares_astro_objects_benchmark.csv")

    ##########################################################
    # Let's define the frame of the alert and the frame to transform to.

    frame_from = "ParticleFrame"
    frame_to = "equatorial"

    ktb.transform_file(
        file0,
        frame_from,
        frame_to,
        detector_from="antares",
        detector_to="antares",
        name="test_toolbox.csv",
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3


def pandas():
    """Imports and returns ``pandas``."""
    try:
        import pandas
    except ImportError:
        raise ImportError(
            "install the 'pandas' package with:\n\n"
            "    pip install pandas\n\n"
            "or\n\n"
            "    conda install pandas"
        )
    else:
        return pandas



"""Constants, like geographical positions."""

import utm


# implied below
HEMISPHERE = "north"
DATUM = "WGS84"


# ORCA
# taken from det ID 49 (km3db CLI -> detx 49)
orca_northing = 4743000
orca_easting = 256500
orca_height = -2440  # m
orca_utm_zone_number = 32
orca_utm_zone_letter = "N"
orca_utm_zone = "{num}{let}".format(num=orca_utm_zone_number, let=orca_utm_zone_letter)
orca_latitude, orca_longitude = utm.to_latlon(
    orca_easting, orca_northing, orca_utm_zone_number, orca_utm_zone_letter
)

# ARCA Sicily site
# taken from det ID 42 (km3db CLI -> detx 49)
arca_northing = 4016800
arca_easting = 587600
arca_height = -3450  # m
arca_utm_zone_number = 33
arca_utm_zone_letter = "N"
arca_utm_zone = "{num}{let}".format(num=arca_utm_zone_number, let=arca_utm_zone_letter)
arca_latitude, arca_longitude = utm.to_latlon(
    arca_easting, arca_northing, arca_utm_zone_number, arca_utm_zone_letter
)

# taken from http://antares.in2p3.fr/internal/dokuwiki/doku.php?id=astro_coordinatetransformation_howto
antares_northing = 4742381.9
antares_easting = 268221.6
antares_height = -2500  # m     (guessed, probably similar to orca)
antares_utm_zone_number = 32
antares_utm_zone_letter = "N"
antares_utm_zone = "{num}{let}".format(
    num=antares_utm_zone_number, let=antares_utm_zone_letter
)
antares_latitude, antares_longitude = utm.to_latlon(
    antares_easting, antares_northing, antares_utm_zone_number, antares_utm_zone_letter
)



from unittest import TestCase

from km3net_testdata import data_path
import km3astro.toolbox as ktb
import km3astro.testing_tools as ktt
import km3astro.coord as kc


class TestToolBox(TestCase):
    def setUp(self):
        self.det = "antares"
        self.date = "2007-10-04"
        self.time = "03:03:03"
        self.phi = 97.07
        self.theta = 135.0
        self.az = 277.07
        self.ze = 45.00
        self.ra = 70.613
        self.dec = -1.852
        self.l = 198.7
        self.b = -29.298
        self.threshold = 0.02

    def test_tool_box_eq_to_utm(self):

        SC_check = ktb.eq_to_utm(self.ra, self.dec, self.date, self.time, self.det)
        SC_true = kc.build_event(
            "UTM", self.date, self.time, self.az, self.ze, "deg", self.det
        )
        sep = ktt.test_Skycoord_separation(SC_true, SC_check)
        if self.threshold < sep:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(sep)
                + " > "
                + str(self.threshold)
            )

    def test_tool_box_eq_to_loc(self):

        SC_check = ktb.eq_to_loc(self.ra, self.dec, self.date, self.time, self.det)
        SC_true = kc.build_event(
            "ParticleFrame", self.date, self.time, self.theta, self.phi, "deg", self.det
        )
        sep = ktt.test_Skycoord_separation(SC_true, SC_check)
        if self.threshold < sep:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(sep)
                + " > "
                + str(self.threshold)
            )

    def test_tool_box_eq_to_gal(self):

        SC_check = ktb.eq_to_gal(self.ra, self.dec, self.date, self.time)
        SC_true = kc.build_event(
            "galactic", self.date, self.time, self.l, self.b, "deg"
        )
        sep = ktt.test_Skycoord_separation(SC_true, SC_check)
        if self.threshold < sep:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(sep)
                + " > "
                + str(self.threshold)
            )

    def test_tool_box_loc_to_eq(self):

        SC_check = ktb.loc_to_eq(self.phi, self.theta, self.date, self.time, self.det)
        SC_true = kc.build_event(
            "equatorial", self.date, self.time, self.ra, self.dec, "deg"
        )
        sep = ktt.test_Skycoord_separation(SC_true, SC_check)
        if self.threshold < sep:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(sep)
                + " > "
                + str(self.threshold)
            )

    def test_tool_box_loc_to_utm(self):

        SC_check = ktb.loc_to_utm(self.phi, self.theta, self.date, self.time, self.det)
        SC_true = kc.build_event(
            "UTM", self.date, self.time, self.az, self.ze, "deg", self.det
        )
        sep = ktt.test_Skycoord_separation(SC_true, SC_check)
        if self.threshold < sep:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(sep)
                + " > "
                + str(self.threshold)
            )

    def test_tool_box_loc_to_gal(self):

        SC_check = ktb.loc_to_gal(self.phi, self.theta, self.date, self.time, self.det)
        SC_true = kc.build_event(
            "galactic", self.date, self.time, self.l, self.b, "deg"
        )
        sep = ktt.test_Skycoord_separation(SC_true, SC_check)
        if self.threshold < sep:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(sep)
                + " > "
                + str(self.threshold)
            )

    def test_get_az_zenith(self):

        SC_true = kc.build_event(
            "UTM", self.date, self.time, self.az, self.ze, "deg", self.det
        )

        az, ze = ktb.get_az_zenith(SC_true)
        if az != self.az or ze != self.ze:
            raise ValueError(
                "Error:  ktb.get_az_zenith return wrong value : "
                + str(az)
                + " != "
                + str(self.az)
                + " and "
                + str(ze)
                + " != "
                + str(self.ze)
            )

    def test_get_phi_theta(self):

        SC_true = kc.build_event(
            "ParticleFrame", self.date, self.time, self.theta, self.phi, "deg", self.det
        )

        phi, theta = ktb.get_phi_theta(SC_true)
        if phi != self.phi or theta != self.theta:
            raise ValueError(
                "Error:  ktb.get_phi_theta return wrong value : "
                + str(phi)
                + " != "
                + str(self.phi)
                + " and "
                + str(theta)
                + " != "
                + str(self.theta)
            )

    def test_get_ra_dec(self):

        SC_true = kc.build_event(
            "equatorial", self.date, self.time, self.ra, self.dec, "deg"
        )

        ra, dec = ktb.get_ra_dec(SC_true)
        if ra != self.ra or dec != self.dec:
            raise ValueError(
                "Error:  ktb.get_ra_dec return wrong value : "
                + str(ra)
                + " != "
                + str(self.ra)
                + " and "
                + str(dec)
                + " != "
                + str(self.dec)
            )

    def test_get_l_b(self):

        SC_true = kc.build_event(
            "galactic", self.date, self.time, self.l, self.b, "deg"
        )

        l, b = ktb.get_l_b(SC_true)
        if l != self.l or b != self.b:
            raise ValueError(
                "Error:  ktb.get_l_b return wrong value : "
                + str(l)
                + " != "
                + str(self.l)
                + " and "
                + str(b)
                + " != "
                + str(self.b)
            )

    def test_get_frame_name(self):

        SC_true = kc.build_event(
            "galactic", self.date, self.time, self.l, self.b, "deg"
        )

        fname = ktb.get_frame_name(SC_true)
        if fname != "galactic":
            raise ValueError(
                "Error:  ktb.get_frame_name return wrong value : "
                + fname
                + " != "
                + "galactic"
            )

    def test_get_obstime(self):

        SC_true = kc.build_event(
            "galactic", self.date, self.time, self.l, self.b, "deg"
        )

        obstime = ktb.get_obstime(SC_true)
        true_obstime = self.date + "T" + self.time
        if obstime != true_obstime:
            raise ValueError(
                "Error:  ktb.get_obstime return wrong value : "
                + obstime
                + " != "
                + true_obstime
            )

    def test_transform_file(self):

        data = ktb.transform_file(
            data_path("astro/antares_coordinate_systems_benchmark.csv"),
            "equatorial",
            "galactic",
            "antares",
            "antares",
            return_data=True,
        )

        l = data["gal_lon"][0]
        b = data["gal_lat"][0]

        max_l = l - self.l
        max_b = b - self.b

        if max_l > self.threshold or max_b > self.threshold:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(max_l)
                + " / "
                + str(max_b)
                + " > "
                + str(self.threshold)
            )

    def test_build_skycoord_list(self):

        table = ktb.reader_from_file(
            data_path("astro/antares_coordinate_systems_benchmark.csv")
        )

        data = ktb.build_skycoord_list(table, "equatorial", "antares")
        table_gal_to_eq = kc.transform_to_new_frame(
            table, "galactic", "equatorial", "antares", "antares"
        )

        sep = ktt.test_benchmark_conversion(data, table_gal_to_eq)

        max_ = sep.max()

        if max_ > self.threshold:
            raise AssertionError(
                "Error: Maximum angle separation = "
                + str(max_)
                + " > "
                + str(self.threshold)
            )



"""Random sampling utilities.
"""
import numpy as np

from km3astro.time import np_to_astrotime


def random_azimuth(n=1, unit="rad"):
    """Draw azimuth, uniformly distributed."""
    n = int(n)  # the method does not cast floats
    azi = 2 * np.pi * np.random.random_sample(size=n)
    if unit == "rad":
        return azi
    elif unit == "deg":
        return azi / np.pi * 180
    else:
        raise KeyError("Unknown unit '{}'".format(unit))


def random_zenith(n=1, unit="rad"):
    """Draw zenith, uniformly distributed in cos(zen)."""
    n = int(n)  # the method does not cast floats
    coszen = 2 * np.random.random_sample(size=n) - 1
    zen = np.arccos(coszen)
    if unit == "rad":
        return zen
    elif unit == "deg":
        return zen / np.pi * 180
    else:
        raise KeyError("Unknown unit '{}'".format(unit))


def second_from_interval(start, stop, n=1, **randargs):
    """Sample random times from an interval (in seconds)."""
    n = int(n)
    sec = np.timedelta64(1, "s")
    n_seconds = (stop - start) / sec
    samples = np.random.randint(low=0, high=n_seconds, size=n, **randargs) * sec
    return start + samples


def equidistant_from_interval(start, stop, step=np.timedelta64(2, "m")):
    """Draw equidistant samples (fixed stepsize) from interval."""
    start = np.datetime64(start)
    stop = np.datetime64(stop)
    duration = stop - start
    n_steps = np.ceil(duration / step)
    samples = np.arange(n_steps) * step
    return start + samples


def random_date(n=1, year=2015, astropy_time=False, **randargs):
    """Create random dates in the given year.

    Parameters
    ----------
    year: int, default: 2015
    n: int, default: 1
    """
    start = np.datetime64("{}-01-01".format(year))
    stop = np.datetime64("{}-01-01".format(year + 1))
    sample = second_from_interval(start, stop, n=n, **randargs)
    if astropy_time:
        return np_to_astrotime(sample)
    else:
        return sample



"""Coordinate transformations.

Galactic:
    GC at (0, 0),
    gal. longitude, latitude (l, b)

Horizontal / altaz (km3):
    centered at detector position
    altitude, azimuth (altitude = 90deg - zenith)

EquatorialJ200 / FK5 / ICRS / GCRS
    (right ascension, declination)

    Equatorial is the same as FK5. FK5 is superseded by the ICRS, so use
    this instead. Note that FK5/ICRS are _barycentric_ implementations,
    so if you are looking for *geocentric* equatorial (i.e.
    for solar system bodies), use GCRS.


A note on maing conventions:
``phi`` and ``theta`` refer to neutrino directions, ``azimuth`` and
``zenith`` to source directions (i.e. the inversed neutrino direction).
The former says where the neutrino points to, the latter says where it comes
from.

Also radian is the default. Degree can be used, but generally the default is
to assume radian.
"""
from astropy import units as u
from astropy.units import rad, deg, hourangle  # noqa
from astropy.coordinates import (
    EarthLocation,
    SkyCoord,
    AltAz,
    Longitude,
    Latitude,
    get_sun,
    get_moon,
)
import astropy.time
from astropy.coordinates import ICRS, Galactic, FK4, FK5  # Low-level frames
from astropy.time import Time
from astropy.coordinates import Angle

import numpy as np
import pandas as pd
import numbers

# also import get_location and convergence_angle that has been shifted to km3frame
import km3astro.frame as kf

from km3astro.constants import (
    arca_longitude,
    arca_latitude,
    arca_height,
    orca_longitude,
    orca_latitude,
    orca_height,
    antares_longitude,
    antares_latitude,
    antares_height,
)
from km3astro.time import np_to_astrotime
from km3astro.random import random_date, random_azimuth
from km3astro.sources import GALACTIC_CENTER


def neutrino_to_source_direction(phi, theta, radian=True):
    """Flip the direction.

    Parameters
    ----------
    phi, theta: neutrino direction
    radian: bool [default=True]
        receive + return angles in radian? (if false, use degree)

    """
    phi = np.atleast_1d(phi).copy()
    theta = np.atleast_1d(theta).copy()
    if not radian:
        phi *= np.pi / 180
        theta *= np.pi / 180
    assert np.all(phi <= 2 * np.pi)
    assert np.all(theta <= np.pi)
    azimuth = (phi + np.pi) % (2 * np.pi)
    zenith = np.pi - theta
    if not radian:
        azimuth *= 180 / np.pi
        zenith *= 180 / np.pi
    return azimuth, zenith


def source_to_neutrino_direction(azimuth, zenith, radian=True):
    """Flip the direction.

    Parameters
    ----------
    zenith : float
        neutrino origin
    azimuth: float
        neutrino origin
    radian: bool [default=True]
        receive + return angles in radian? (if false, use degree)

    """
    azimuth = np.atleast_1d(azimuth).copy()
    zenith = np.atleast_1d(zenith).copy()
    if not radian:
        azimuth *= np.pi / 180
        zenith *= np.pi / 180
    phi = (azimuth - np.pi) % (2 * np.pi)
    theta = np.pi - zenith
    if not radian:
        phi *= 180 / np.pi
        theta *= 180 / np.pi
    return phi, theta


def Sun(time):
    """Wrapper around astropy's get_sun, accepting numpy/pandas time objects."""
    if not isinstance(time, astropy.time.Time):
        # if np.datetime64, convert to astro time
        time = np_to_astrotime(time)
    return get_sun(time)


def Moon(time):
    """Wrapper around astropy's get_moon, accepting numpy/pandas time objects."""
    if not isinstance(time, astropy.time.Time):
        # if np.datetime64, convert to astro time
        time = np_to_astrotime(time)
    return get_moon(time)


def local_frame(time, location):
    """Get the (horizontal) coordinate frame of your detector."""
    if not isinstance(time, astropy.time.Time):
        # if np.datetime64, convert to astro time
        time = np_to_astrotime(time)
    loc = kf.get_location(location)
    frame = AltAz(obstime=time, location=loc)
    return frame


def local_event(azimuth, time, zenith, location, radian=True, **kwargs):
    """Create astropy events from detector coordinates."""
    zenith = np.atleast_1d(zenith).copy()
    azimuth = np.atleast_1d(azimuth).copy()
    if not radian:
        azimuth *= np.pi / 180
        zenith *= np.pi / 180
    altitude = zenith - np.pi / 2

    loc = kf.get_location(location)
    # neutrino telescopes call the co-azimuth "azimuth"
    true_azimuth = (
        np.pi / 2 - azimuth + np.pi + kf.convergence_angle(loc.lat.rad, loc.lon.rad)
    ) % (2 * np.pi)
    frame = local_frame(time, location=location)
    event = SkyCoord(alt=altitude * rad, az=true_azimuth * rad, frame=frame, **kwargs)
    return event


def sun_local(time, loc):
    """Sun position in local coordinates."""
    frame = local_frame(time, location=loc)
    sun = Sun(time)
    sun_local = sun.transform_to(frame)
    return sun_local


def moon_local(time, loc):
    """Moon position in local coordinates."""
    frame = local_frame(time, location=loc)
    moon = Moon(time)
    moon_local = moon.transform_to(frame)
    return moon_local


def gc_in_local(time, loc):
    """Galactic center position in local coordinates."""
    frame = local_frame(time, location=loc)
    gc = GALACTIC_CENTER
    gc_local = gc.transform_to(frame)
    return gc_local


def orca_gc_dist(azimuth, time, zenith, frame="detector"):
    """Return angular distance of event to GC.

    Parameters
    ==========
    frame: str, [default: 'detector']
        valid are 'detector', 'galactic', 'icrs', 'gcrs'
    """
    evt = local_event(azimuth, time, zenith)
    galcen = gc_in_local(time, loc)
    if frame == "detector":
        pass
    elif frame in ("galactic", "icrs", "gcrs"):
        evt = evt.transform_to(frame)
        galcen = galcen.transform_to(frame)
    return evt.separation(galcen).radian


def orca_sun_dist(azimuth, time, zenith):
    """Return distance of event to sun, in detector coordinates."""
    evt = local_event(azimuth, time, zenith)
    sun = sun_local(time, loc)
    dist = evt.separation(sun).radian
    return dist


def gc_dist_random(zenith, frame="detector"):
    """Generate random (time, azimuth) events and get distance to GC."""
    n_evts = len(zenith)
    time = random_date(n=n_evts)
    azimuth = random_azimuth(n=n_evts)
    dist = orca_gc_dist(azimuth, time, zenith, frame=frame)
    return dist


def sun_dist_random(zenith):
    """Generate random (time, azimuth) events and get distance to GC."""
    n_evts = len(zenith)
    time = random_date(n=n_evts)
    azimuth = random_azimuth(n=n_evts)
    dist = orca_sun_dist(azimuth, time, zenith)
    return dist


class Event(object):
    def __init__(self, zenith, azimuth, time, location):
        self.zenith = zenith
        self.azimuth = azimuth
        self.time = time

    @classmethod
    def from_zenith(cls, zenith, **initargs):
        zenith = np.atleast_1d(zenith)
        n_evts = zenith.shape[0]
        azimuth = random_azimuth(n_evts)
        time = random_date(n_evts)
        return cls(zenith, azimuth, time, **initargs)


def is_args_fine_for_frame(frame, *args):

    if frame == "ParticleFrame" and len(args) < 6:
        raise TypeError(
            "Only "
            + str(len(args))
            + " given when 6 are needed: date, time, theta, phi, unit, particleframe ! for ParticleFrame"
        )

    if frame == "UTM" and len(args) < 6:
        raise TypeError(
            "Only "
            + str(len(args))
            + " given when 6 are needed: date, time, azimuth, zenith, unit, particleframe ! for UTM"
        )

    if frame == "equatorial" and len(args) < 4:
        raise TypeError(
            "Only "
            + str(len(args))
            + " given when 4 are needed: date, time, ra, dec ! for Equatorial"
        )

    if frame == "galactic" and len(args) < 4:
        raise TypeError(
            "Only "
            + str(len(args))
            + " given when 4 are needed: date, time, l, b ! for Galactic"
        )

    return 0


def build_event(Cframe, *args):
    """Build a SkyCoord object of the corresponding frame and parameters

    Parameters
    ----------
    Cframe : str
        Frame of the Skycoord event to build, either "ParticleFrame", "UTM", "equatorial" or "galactic"
    *args : list of the sky coordinate parameters

    Returns
    -------
    SkyCoord : astropy.SkyCoord
        Sky coordinate object

    """

    is_args_fine_for_frame(Cframe, *args)

    # Defining time of observation
    # using astropy.time Time because pandas.to_datetime raise an error for convertion in np_to_astrotime
    time = args[0] + "T" + args[1]
    time = Time(time)

    # ParticleFrame : date, time , theta, phi, unit, detector_name
    if Cframe == "ParticleFrame":

        theta = args[2]
        phi = args[3]
        unit = args[4]
        if unit == "deg":
            phi = phi * u.deg
            theta = theta * u.deg

        else:
            phi = phi * u.rad
            theta = theta * u.rad

        loc = kf.get_location(args[5])
        r = u.Quantity(100, u.m)  # dummy r value ! Warning !
        return SkyCoord(
            frame=kf.ParticleFrame,
            phi=phi,
            theta=theta,
            location=loc,
            obstime=time,
            r=r,
        )

    # UTM : date, time, azimuth, zenith, unit, detector
    elif Cframe == "UTM":

        az = args[2]
        zenith = args[3]
        unit = args[4]
        if unit == "deg":
            az = az * u.deg
            zenith = zenith * u.deg

        else:
            az = az * u.rad
            zenith = zenith * u.rad

        loc = kf.get_location(args[5])
        r = u.Quantity(100, u.m)  # dummy r value ! Warning !

        return SkyCoord(
            frame=kf.UTM, azimuth=az, zenith=zenith, location=loc, obstime=time, r=r
        )

    elif Cframe == "galactic":
        l = args[2]
        b = args[3]

        if isinstance(l, str):
            l = Angle(l, unit="hourangle")
        elif isinstance(l, numbers.Number):
            l = Angle(l, unit=u.deg)

        if isinstance(b, str):
            b = Angle(b, unit="hourangle")
        elif isinstance(b, numbers.Number):
            b = Angle(b, unit=u.deg)

        return SkyCoord(frame=Cframe, l=l, b=b, unit="deg", obstime=time)

    elif Cframe == "equatorial":
        ra = args[2]
        dec = args[3]

        if isinstance(ra, str):
            ra = Angle(ra, unit="hourangle")
        elif isinstance(ra, numbers.Number):
            ra = Angle(ra, unit=u.deg)

        if isinstance(dec, str):
            dec = Angle(dec, unit=u.deg)
        elif isinstance(dec, numbers.Number):
            dec = Angle(dec, unit=u.deg)

        return SkyCoord(frame=ICRS, ra=ra, dec=dec, obstime=time)

    else:
        raise ValueError("Error: Wrong Frame input:" + Cframe)
        return None


def transform_to(Skycoord, frame_to, detector_to="antares"):
    """Transform a Skycoord object to the desired frame

    Parameters
    ----------
    SkyCoord : astropy.SkyCoord
        The sky coordinate
    frame_to : str
        The desired frame of transformation, either "ParticleFrame", "UTM", "altaz", "equatorial or "galactic"
    detector_to : str [default = "antares"]
        The detector of the transformed frame, either "orca", "arca" or "antares"

    """

    time = Skycoord.obstime
    loc = kf.get_location(detector_to)

    if frame_to == "ParticleFrame":

        frame = kf.ParticleFrame(obstime=time, location=loc)
        return Skycoord.transform_to(frame)

    elif frame_to == "UTM":
        frame = kf.UTM(obstime=time, location=loc)
        return Skycoord.transform_to(frame)

    elif frame_to == "altaz":
        frame = AltAz(obstime=time, location=loc)
        return Skycoord.transform_to(frame)

    elif frame_to == "equatorial":
        return Skycoord.transform_to(ICRS)

    elif frame_to == "galactic":
        return Skycoord.transform_to("galactic")

    else:
        raise ValueError("Wrong Frame to transform: " + frame_to + " is not valid")
        return -1


def transform_to_new_frame(
    table, frame, frame_to, detector="antares", detector_to="antares"
):
    """Transform a pandas DataFrame of sky coordinate parameters to a new pandas.DataFrame of SkyCoord of the initial and desired frame

    Parameters
    ----------
    table : pandas.DataFrame(astropy.SkyCoord)
        The sky coordinate
    frame : str
        The frame of the table of parameters, either "ParticleFrame", "UTM", "altaz", "equatorial or "galactic"
    detector : str [default = "antares"]
        The detector of the sky coordinate parameters, either "orca", "arca" or "antares"
    frame_to : str
        The desired frame of transformation, either "ParticleFrame", "UTM", "altaz", "equatorial or "galactic"
    detector_to : str [default = "antares"]
        The detector of the transformed frame, either "orca", "arca" or "antares"

    """

    if frame == "ParticleFrame":
        list_evt = table.apply(
            lambda x: build_event(
                frame, x.date, x.time, x.theta, x.phi, "deg", detector
            ),
            axis=1,
            result_type="expand",
        )

    if frame == "UTM":
        list_evt = table.apply(
            lambda x: build_event(
                frame, x.date, x.time, x.azimuth, x.zenith, "deg", detector
            ),
            axis=1,
            result_type="expand",
        )

    if frame == "equatorial":
        list_evt = table.apply(
            lambda x: build_event(frame, x.date, x.time, x["RA-J2000"], x["DEC-J2000"]),
            axis=1,
            result_type="expand",
        )

    if frame == "galactic":
        list_evt = table.apply(
            lambda x: build_event(frame, x.date, x.time, x.gal_lon, x.gal_lat),
            axis=1,
            result_type="expand",
        )

    if isinstance(list_evt, pd.Series):
        series = {"SkyCoord_base": list_evt}
        list_evt = pd.DataFrame(series)

    list_evt.set_axis(["SkyCoord_base"], axis="columns", inplace=True)

    list_evt["SkyCoord_new"] = list_evt.apply(
        lambda x: transform_to(x.SkyCoord_base, frame_to, detector_to),
        axis=1,
        result_type="expand",
    )

    return list_evt



from astropy.units import hourangle, deg
from astropy.coordinates import SkyCoord

# (right ascension, declination)
SIRIUS = SkyCoord("06 45 08.9 -16 42 58", unit=(hourangle, deg))
CANOPUS = SkyCoord("06 23 57.1 -52 41 45", unit=(hourangle, deg))
ARCTURUS = SkyCoord("14 15 39.7 +19 10 57", unit=(hourangle, deg))
ANTARES = SkyCoord("16 29 24.4 -26 25 55", unit=(hourangle, deg))

RX_J1713 = SkyCoord("17 17 33.6 -39 45 36.4", unit=(hourangle, deg))
VELA_X = SkyCoord("08 35 00 -45 36 00", unit=(hourangle, deg))

# Sagittarius A* (sagittarius a-star)
# GALACTIC_CENTER = SkyCoord(0 * deg, 0 * deg, frame='galactic').icrs
SAGITTARIUS_A_STAR = SkyCoord(
    "17 45 40.0409 -29 0 28.118", unit=(hourangle, deg)
)  # noqa
GALACTIC_CENTER = SAGITTARIUS_A_STAR



from unittest import TestCase

from km3astro.random import random_azimuth, random_zenith, random_date


class TestRandom(TestCase):
    def setUp(self):
        self.n_evts = 100
        self.n_evts_funny = 1e2

    def test_zenith(self):
        zen = random_zenith(n=self.n_evts)
        assert zen.shape[0] == self.n_evts
        zen2 = random_zenith(n=self.n_evts_funny)
        self.assertAlmostEqual(zen2.shape[0], self.n_evts_funny)

    def test_azimuth(self):
        azi = random_azimuth(n=self.n_evts_funny)
        assert azi.shape[0] == self.n_evts_funny
        azi2 = random_azimuth(n=self.n_evts_funny)
        self.assertAlmostEqual(azi2.shape[0], self.n_evts_funny)

    def test_date(self):
        tim = random_date(n=self.n_evts)
        assert tim.shape[0] == self.n_evts
        tim2 = random_date(n=self.n_evts_funny)
        self.assertAlmostEqual(tim2.shape[0], self.n_evts_funny)



"""
========================
Sun in local coordinates
========================

Show off some coordinate transformations.
"""

# Author: Moritz Lotze <mlotze@km3net.de>
# License: BSD-3

from astropy.units import deg
import numpy as np
import pandas as pd

from km3astro.random import random_date, random_azimuth, random_zenith
from km3astro.coord import local_frame, Sun, source_to_neutrino_direction


##########################################################
# generate some random events

n_evts = 1e4
zen = random_zenith(n=n_evts)
time = random_date(n=n_evts)
azi = random_azimuth(n=n_evts)

##########################################################
# transform to horizontal coordinates

orca_frame = local_frame(time=time, location="orca")
sun = Sun(time)

sun_orca = sun.transform_to(orca_frame)

sun_azi = sun_orca.az.rad
sun_zen = (90 * deg - sun_orca.alt).rad

sun_phi, sun_theta = source_to_neutrino_direction(sun_azi, sun_zen)

sun_df = pd.DataFrame(
    {
        "Sun Azimuth": sun_azi,
        "Sun Zenith": sun_zen,
        "Sun Cos Zenith": np.cos(sun_zen),
        "Sun Phi": sun_phi,
        "Sun Theta": sun_theta,
        "Sun Cos Theta": np.cos(sun_theta),
    }
)

#########################################################

sun_df.plot.hexbin("Sun Zenith", "Sun Azimuth", cmap="viridis")


#########################################################

sun_df.plot.hexbin("Sun Cos Zenith", "Sun Azimuth", cmap="magma")

#########################################################

sun_df.plot.hexbin("Sun Theta", "Sun Phi", cmap="inferno")

#########################################################

sun_df.plot.hexbin("Sun Cos Theta", "Sun Phi", cmap="plasma")



from unittest import TestCase

import numpy as np
import pandas as pd

from km3astro.time import np_to_datetime
from km3astro.random import (
    second_from_interval,
    equidistant_from_interval,
    random_date,
)


class TestTime(TestCase):
    def test_np_datetime(self):
        cur = np.datetime64("2017-03-07T15:05:54.117605")
        assert np.alltrue(cur == np_to_datetime([cur]))
        assert np.alltrue(cur == np_to_datetime(cur))

    def test_rand(self):
        randtime = random_date(n=100)
        np_to_datetime(randtime)

    def test_dataframe(self):
        df = pd.DataFrame({"time": random_date(n=100)})
        np_to_datetime(df.time)



"""Time conversion utilities.

For random time samples, goto `km3flux.random`
"""
from datetime import datetime

import numpy as np
from astropy.time import Time


def now():
    """Returns the time of now as astropy Time"""
    return Time(datetime.utcnow())


def np_to_datetime(intime):
    """Convert numpy/pandas datetime64 to list[datetime]."""
    nptime = np.atleast_1d(intime)
    np_corr = (nptime - np.datetime64("1970-01-01T00:00:00")) / np.timedelta64(1, "s")
    return [datetime.utcfromtimestamp(t) for t in np_corr]


def np_to_astrotime(intime):
    return Time(np_to_datetime(intime))



from unittest import TestCase

from km3astro.constants import (
    orca_easting,
    orca_northing,
    orca_utm_zone_letter,
    orca_utm_zone_number,
    orca_latitude,
    orca_longitude,
    arca_easting,
    arca_northing,
    arca_utm_zone_letter,
    arca_utm_zone_number,
    arca_latitude,
    arca_longitude,
)

# taken from loi
arca_latitude_naive = 36 + (16 / 60)  # degree
arca_longitude_naive = 16 + (6 / 60)  # degree
arca_lat_delta = 0.3
arca_lon_delta = 0.3

orca_latitude_naive = 42 + (48.05 / 60)  # degree
orca_longitude_naive = 6 + (1.3368 / 60)  # degree
orca_lat_delta = 0.01
orca_lon_delta = 0.01


class TestUTM(TestCase):
    def test_ref_vs_computed(self):
        self.assertAlmostEqual(
            orca_longitude_naive, orca_longitude, delta=orca_lon_delta
        )
        self.assertAlmostEqual(orca_latitude_naive, orca_latitude, delta=orca_lat_delta)
        self.assertAlmostEqual(
            arca_longitude_naive, arca_longitude, delta=arca_lon_delta
        )
        self.assertAlmostEqual(arca_latitude_naive, arca_latitude, delta=arca_lat_delta)



"""
======================================================
Benchmark Calculations for Coordinates Transformations
======================================================

Run some benchmarks similar to
`http://antares.in2p3.fr/internal/dokuwiki/doku.php?id=benchmarks_astro`

CAVEAT:
    IceCube and ANTARES/KM3NeT define "azimuth" differently:

``true_azimuth = (90 - azimuth) % 360``!

"""

# Author: Moritz Lotze <mlotze@km3net.de>
# License: BSD-3

from collections import OrderedDict

import numpy as np
import pandas as pd

from km3astro.coord import local_event, local_frame, neutrino_to_source_direction
from km3astro.sources import SIRIUS, CANOPUS, ARCTURUS, ANTARES  # noqa


time = pd.to_datetime(
    [
        "2007-10-04 03:03:03.00",
        "2007-10-04 03:03:03.00",
        "2007-10-04 03:03:03.00",
        "2007-10-04 03:03:03.00",
        "2007-10-04 03:03:03.00",
    ]
)
theta_deg = np.array(
    [
        135.00,
        11.97,
        22.97,
        33.97,
        85.23,
    ]
)
phi_deg = np.array(
    [
        97.07,
        23.46,
        97.07,
        192.50,
        333.33,
    ]
)
theta = theta_deg * np.pi / 180
phi = phi_deg * np.pi / 180
azimuth, zenith = neutrino_to_source_direction(phi, theta)

zenith = np.pi - zenith
azimuth = np.pi + azimuth

evt = local_event(azimuth, time, zenith, location="antares")
equat = evt.fk5
# dec = equat.dec.degree
# ra = equat.ra.degree
dec = equat.dec
ra = equat.ra
gal = evt.galactic
l = gal.l
b = gal.b

data = pd.DataFrame(
    OrderedDict(
        [
            ("time", time),
            ("theta [deg]", theta_deg),
            ("phi [deg]", phi_deg),
            ("theta", theta),
            ("phi", phi),
            ("zenith [deg]", zenith * 180 / np.pi),
            ("azimuth [deg]", azimuth * 180 / np.pi),
            ("zenith", zenith),
            ("azimuth", azimuth),
            ("declication", dec),
            ("right ascension", ra),
            ("gal_longitude", l),
            ("gal_latitude", b),
        ]
    )
)
print(data[:])

########################################################
# look at some sources in horizontal coordinates

frame = local_frame(
    time=pd.to_datetime(
        [
            "2007-10-04 03:03:03.00",
        ]
    ),
    location="antares",
)
sirius_local = SIRIUS.transform_to(frame)
sirius_alt = sirius_local.alt.degree
sirius_az = sirius_local.az.degree
sirius_zen = 90 - sirius_alt
print()
print(sirius_az)
print(sirius_zen)

canopus_local = CANOPUS.transform_to(frame)
canopus_alt = canopus_local.alt.degree
canopus_az = canopus_local.az.degree
canopus_zen = 90 - canopus_alt
print()
print(canopus_az)
print(canopus_zen)



from unittest import TestCase

from km3net_testdata import data_path
import km3astro.testing_tools as ktt


class TestAntaresBenchmarks(TestCase):
    def setUp(self):
        self.angle_threshold = 0.02

    def test_antares_objects(self):
        ktt.test_angle_separation(
            data_path("astro/antares_astro_objects_benchmark.csv"), "antares", "antares"
        )

    def test_antares_coordinate_system_benchmarks(self):
        ktt.test_angle_separation(
            data_path("astro/antares_coordinate_systems_benchmark.csv"),
            "antares",
            "antares",
        )


class TestARCABenchmarks(TestCase):
    def setUp(self):
        self.angle_threshold = 0.02

    def test_arca_objects(self):
        ktt.test_angle_separation(
            data_path("astro/ARCA_astro_objects_benchmark.csv"), "arca", "arca"
        )

    def test_arca_coordinate_system_benchmarks(self):
        ktt.test_angle_separation(
            data_path("astro/ARCA_coordinate_systems_benchmark.csv"), "arca", "arca"
        )


class TestORCABenchmarks(TestCase):
    def setUp(self):
        self.angle_threshold = 0.02

    def test_orca_objects(self):
        ktt.test_angle_separation(
            data_path("astro/ORCA_astro_objects_benchmark.csv"), "orca", "orca"
        )

    def test_orca_coordinate_system_benchmarks(self):
        ktt.test_angle_separation(
            data_path("astro/ORCA_coordinate_systems_benchmark.csv"), "orca", "orca"
        )



#!/usr/bin/env python
# Filename: setup.py
"""
km3astro setup script.

"""
import os
from setuptools import setup
import sys


def read_requirements(kind):
    """Return a list of stripped lines from a file"""
    with open(os.path.join("requirements", kind + ".txt")) as fobj:
        requirements = [l.strip() for l in fobj.readlines()]
    v = sys.version_info
    if (v.major, v.minor) < (3, 6):
        try:
            requirements.pop(requirements.index("black"))
        except ValueError:
            pass
    return requirements


try:
    with open("README.rst") as fh:
        long_description = fh.read()
except UnicodeDecodeError:
    long_description = "km3astro - astronomical utilities for KM3NeT"

setup(
    name="km3astro",
    url="http://git.km3net.de/km3py/km3astro",
    description="Astronomical utilities for KM3NeT",
    long_description=long_description,
    author="Tamas Gal and Moritz Lotze",
    author_email="tgal@km3net.de",
    packages=["km3astro"],
    include_package_data=True,
    platforms="any",
    setup_requires=["numpy>=1.12", "setuptools_scm"],
    install_requires=read_requirements("install"),
    extras_require={kind: read_requirements(kind) for kind in ["dev", "extras"]},
    use_scm_version=True,
    python_requires=">=3.5",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
    ],
)

__author__ = "Tamas Gal and Moritz Lotze"



"""
===============================
Local to Equatorial Coordinates
===============================

Where do my neutrinos come from?
"""

# sphinx_gallery_thumbnail_number = 3

__author__ = "moritz"

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from km3astro.coord import local_event, neutrino_to_source_direction
from km3astro.frame import get_location
from km3astro.plot import plot_equatorial
from km3astro.sources import VELA_X


##########################################################
# Detector Coordinates
# --------------------
# Let's define some random events.

theta = 10 * np.pi / 180
phi = 8 * np.pi / 180
time = pd.to_datetime(
    [
        "2015-01-12T15:10:12",
        "2015-03-15T14:24:01",
    ]
).values[0]

##########################################################
# Phi, theta: Where the neutrino is pointing to
#
# Zenith, azimuth: where the neutrino is coming from

azimuth, zenith = neutrino_to_source_direction(phi, theta, radian=True)

#########################################################################
# We want to observe them from the Orca location. Let's look at our
# geographical coordinates.
#
# In km3astro, there are the predefined locations "orca", "arca" and "antares".
orca_loc = get_location("orca")

#########################################################################
# Create event in local coordinates (aka AltAz or Horizontal Coordinates)
#
# This returns an ``astropy.SkyCoord`` instance.

evt_local = local_event(azimuth=azimuth, zenith=zenith, time=time, location="orca")

print(evt_local)

##############################################################
# Transform to equatorial -- ICRS
# -------------------------------
#
# "If you’re looking for “J2000” coordinates, and aren’t sure if
# you want to use this or FK5, you probably want to use ICRS. It’s more
# well-defined as a catalog coordinate and is an inertial system, and is
# very close (within tens of milliarcseconds) to J2000 equatorial."

evt_equat = evt_local.transform_to("icrs")
print(evt_equat)

##############################################################
# How far removed are these events from a certain source?

source_dist = evt_equat.separation(VELA_X)

plt.hist(source_dist.degree, bins="auto")

##############################################################
# Plot them in a square

right_ascension_radian = evt_equat.ra.rad
declination_radian = evt_equat.dec.rad

plt.scatter(right_ascension_radian, declination_radian)
plt.scatter(VELA_X.ra.rad, VELA_X.dec.rad)
plt.xlabel("Right Ascension / rad")
plt.ylabel("Declination / rad")

##############################################################
# Plot them in a skymap.
#
# We need this little wrap because astropy's
# convention for ra, dec differs from matplotlib.

ax = plot_equatorial(evt_equat, markersize=12, label="Event")
plot_equatorial(VELA_X, markersize=12, ax=ax, label="Vela X")
plt.legend(loc="best")



"""
========================
Multiple exemple of how to use km3astro.plot
========================

"""

# Author: Tedjditi Hichem <htedjditi@km3net.de>


import numpy as np
import pandas as pd

import km3astro.plot as kp
from km3net_testdata import data_path


def main():

    fig = kp.skymap_list(
        data_path("astro/antares_coordinate_systems_benchmark.csv"),
        frame="UTM",
        detector="antares",
        plot_frame="equatorial",
        detector_to="antares",
        save=False,
    )

    # fig.savefig("test_save_fig.png")

    table_read = pd.read_csv(
        data_path("astro/antares_coordinate_systems_benchmark.csv"), comment="#"
    )

    alert_type = [
        "GRB",
        "GW",
        "Neutrino",
        "NuEM",
        "SK_SN",
        "SNEWS",
        "Transient",
        "Random",
        "GRB",
        "GW",
        "Neutrino",
        "NuEM",
        "SK_SN",
        "SNEWS",
        "Transient",
        "Randome",
        "Hasard",
    ]

    table_read["Alert_type"] = alert_type

    _ = kp.skymap_list(
        dataframe=table_read,
        frame="UTM",
        detector="antares",
        plot_frame="equatorial",
        detector_to="antares",
        title="test_title_input",
        save=True,
        name="test_dataframe_input",
    )

    _ = kp.skymap_list(
        data_path("astro/antares_coordinate_systems_benchmark.csv"),
        frame="UTM",
        detector="antares",
        plot_frame="galactic",
        detector_to="antares",
    )

    _ = kp.skymap_list(
        data_path("astro/ORCA_coordinate_systems_benchmark.csv"),
        frame="UTM",
        detector="orca",
        plot_frame="equatorial",
        detector_to="orca",
    )

    _ = kp.skymap_list(
        data_path("astro/ORCA_coordinate_systems_benchmark.csv"),
        frame="UTM",
        detector="orca",
        plot_frame="galactic",
        detector_to="orca",
    )

    _ = kp.skymap_list(
        data_path("astro/ARCA_coordinate_systems_benchmark.csv"),
        frame="UTM",
        detector="arca",
        plot_frame="equatorial",
        detector_to="arca",
    )

    _ = kp.skymap_list(
        data_path("astro/ARCA_coordinate_systems_benchmark.csv"),
        frame="UTM",
        detector="arca",
        plot_frame="galactic",
        detector_to="arca",
    )

    _ = kp.skymap_alert(
        file0=data_path("astro/antares_coordinate_systems_benchmark.csv"),
        frame="UTM",
        detector="antares",
        plot_frame="ParticleFrame",
        detector_to="antares",
    )

    _ = kp.skymap_alert(
        ra=80,
        dec=-20,
        obstime="2022-07-18T03:03:03",
        plot_frame="galactic",
        detector="dummy",
        detector_to="orca",
    )

    _ = kp.skymap_alert(
        ra=80,
        dec=-20,
        obstime="2022-07-18T03:03:03",
        plot_frame="equatorial",
        detector="dummy",
        detector_to="orca",
    )


if __name__ == "__main__":
    main()



from unittest import TestCase

import pandas as pd
from km3net_testdata import data_path
import km3astro.toolbox as ktb
import km3astro.testing_tools as ktt
import km3astro.coord as kc
import km3astro.plot as kp


class TestPlotSkymap(TestCase):
    def test_skymap_list(self):

        _ = kp.skymap_list(
            data_path("astro/antares_coordinate_systems_benchmark.csv"),
            frame="UTM",
            detector="antares",
            plot_frame="galactic",
            detector_to="antares",
            save=True,
        )

        table_read = pd.read_csv(
            data_path("astro/antares_coordinate_systems_benchmark.csv"), comment="#"
        )
        alert_type = [
            "GRB",
            "GW",
            "Neutrino",
            "NuEM",
            "SK_SN",
            "SNEWS",
            "Transient",
            "Random",
            "GRB",
            "GW",
            "Neutrino",
            "NuEM",
            "SK_SN",
            "SNEWS",
            "Transient",
            "Randome",
            "Hasard",
        ]

        table_read["Alert_type"] = alert_type

        _ = kp.skymap_list(
            dataframe=table_read,
            frame="UTM",
            detector="antares",
            plot_frame="equatorial",
            detector_to="antares",
            title="test_title",
            save=True,
            name="test_dataframe_input",
        )

        _ = kp.skymap_list(
            data_path("astro/ORCA_coordinate_systems_benchmark.csv"),
            frame="UTM",
            detector="orca",
            plot_frame="equatorial",
            detector_to="orca",
        )

        _ = kp.skymap_list(
            data_path("astro/ORCA_coordinate_systems_benchmark.csv"),
            frame="UTM",
            detector="orca",
            plot_frame="galactic",
            detector_to="orca",
        )

        _ = kp.skymap_list(
            data_path("astro/ARCA_coordinate_systems_benchmark.csv"),
            frame="UTM",
            detector="arca",
            plot_frame="equatorial",
            detector_to="arca",
        )

        _ = kp.skymap_list(
            data_path("astro/ARCA_coordinate_systems_benchmark.csv"),
            frame="UTM",
            detector="arca",
            plot_frame="galactic",
            detector_to="arca",
        )

    def test_skymap_alert(self):

        _ = kp.skymap_alert(
            file0=data_path("astro/antares_coordinate_systems_benchmark.csv"),
            frame="UTM",
            detector="antares",
            plot_frame="ParticleFrame",
            detector_to="antares",
            save=True,
        )

        _ = kp.skymap_alert(
            ra=80,
            dec=-20,
            obstime="2022-07-18T03:03:03",
            plot_frame="galactic",
            detector="dummy",
            detector_to="orca",
        )

        _ = kp.skymap_alert(
            ra=80,
            dec=-20,
            obstime="2022-07-18T03:03:03",
            plot_frame="equatorial",
            detector="dummy",
            detector_to="orca",
        )

    def test_skymap_hpx(self):

        _ = kp.skymap_hpx(
            file0="https://gracedb.ligo.org/api/superevents/MS210525n/files/bayestar.fits.gz,0",
            save=True,
        )



import numpy as np

from astropy.coordinates import ICRS, Galactic, FK4, FK5  # Low-level frames
import astropy.units as u
from astropy.units import rad, deg, hourangle  # noqa

from astropy.coordinates import Angle

from astropy.coordinates import (
    EarthLocation,
    SkyCoord,
    AltAz,
    Longitude,
    Latitude,
)

from km3astro.constants import (
    arca_longitude,
    arca_latitude,
    arca_height,
    orca_longitude,
    orca_latitude,
    orca_height,
    antares_longitude,
    antares_latitude,
    antares_height,
)


from astropy.coordinates import BaseCoordinateFrame
import astropy.coordinates.representation as rep
from astropy.coordinates import RepresentationMapping
from astropy.coordinates.attributes import (
    TimeAttribute,
    EarthLocationAttribute,
    QuantityAttribute,
)
from astropy.coordinates import TransformGraph, frame_transform_graph, FunctionTransform


LOCATIONS = {
    "arca": EarthLocation.from_geodetic(
        lon=Longitude(arca_longitude * deg),
        lat=Latitude(arca_latitude * deg),
        height=arca_height,
    ),
    "orca": EarthLocation.from_geodetic(
        lon=Longitude(orca_longitude * deg),
        lat=Latitude(orca_latitude * deg),
        height=orca_height,
    ),
    "antares": EarthLocation.from_geodetic(
        lon=Longitude(antares_longitude * deg),
        lat=Latitude(antares_latitude * deg),
        height=antares_height,
    ),
}


def get_location(location):
    try:
        loc = LOCATIONS[location]
    except KeyError:
        raise KeyError("Invalid location, valid are 'orca', 'arca', 'antares'")
    return loc


def convergence_angle(lat, lon):
    """Calculate the converge angle on the UTM grid.

    Parameters
    ----------
    lon : number
        Longitude in rad
    lat : number
        Latitude in rad

    """
    latitude_deg = lat * u.deg

    if latitude_deg > 84 * u.deg or latitude_deg < -80 * u.deg:
        raise ValueError(
            "UTM coordinate system is only defined between -80deg S and 84deg N."
        )

    # detector position, longitude and latitude in rad
    # lambda  = longitude
    phi = lat

    # find UTM zone and central meridian

    # longitude of the central meridian of UTM zone in rad
    lambda0 = longitude_of_central_meridian(utm_zone(lon))
    omega = lon - lambda0

    # parameters of the Earth ellipsoid
    sma = 6378137  # semi-major axis in meters (WGS84)
    ecc = 0.0066943800  # eccentricity (WGS84)

    rho = sma * (1 - ecc) / pow(1 - ecc * np.sin(phi) ** 2, 3 / 2)
    nu = sma / np.sqrt(1 - ecc * np.sin(phi) ** 2)
    psi = nu / rho
    t = np.tan(phi)

    angle = (
        np.sin(phi) * omega
        - np.sin(phi) * omega**3 / 3 * pow(np.cos(phi), 2) * (2 * psi**2 - psi)
        - np.sin(phi)
        * omega**5
        / 15
        * pow(np.cos(phi), 4)
        * (
            psi**4 * (11 - 24 * t**2)
            - psi**3 * (11 - 36 * t**2)
            + 2 * psi**2 * (1 - 7 * t**2)
            + psi * t**2
        )
        - np.sin(phi)
        * omega**7
        / 315
        * pow(np.cos(phi), 6)
        * (17 - 26 * t**2 + 2 * t**4)
    )

    return angle


def utm_zone(lat):
    """The UTM zone for a given latitude

    Parameters
    ----------
    lat : number
        Latitude in rad

    """
    return 1 + int((np.pi + lat) / (6 * np.pi / 180))


def longitude_of_central_meridian(utmzone):
    """The longitude of the central meridian for a given UTM zone.

    Parameters
    ----------
    utmzone : number
        The UTM zone.

    """
    zone_width = 6 * np.pi / 180
    return -np.pi + (utmzone - 1) * zone_width + zone_width / 2


class ParticleFrame(BaseCoordinateFrame):
    default_representation = rep.PhysicsSphericalRepresentation

    # Specify frame attributes required to fully specify the frame
    obstime = TimeAttribute(default=None)
    location = EarthLocationAttribute(default=get_location("arca"))


class UTM(BaseCoordinateFrame):

    default_representation = rep.PhysicsSphericalRepresentation

    frame_specific_representation_info = {
        rep.PhysicsSphericalRepresentation: [
            RepresentationMapping("phi", "azimuth"),
            RepresentationMapping("theta", "zenith"),
        ]
    }

    obstime = TimeAttribute(default=None)
    location = EarthLocationAttribute(default=get_location("arca"))

    @property
    def alt_utm(self):
        """Altitude in UTM"""
        return Angle(np.pi / 2 - self.zenith.rad, u.rad)


@frame_transform_graph.transform(FunctionTransform, ParticleFrame, AltAz)
def ParticleFrame_to_AltAz(ParticleFrame_, altaz):

    phi = ParticleFrame_.phi.rad
    theta = ParticleFrame_.theta.rad
    loc = ParticleFrame_.location
    time = ParticleFrame_.obstime

    conv_angle = Angle(convergence_angle(loc.lat.rad, loc.lon.rad), unit=u.radian)

    altitude = theta - np.pi / 2
    Corrected_azimuth = (np.pi / 2 - phi + np.pi + conv_angle.rad) % (2 * np.pi)

    altaz = AltAz(
        alt=altitude * rad, az=Corrected_azimuth * rad, obstime=time, location=loc
    )

    return altaz


@frame_transform_graph.transform(FunctionTransform, AltAz, ParticleFrame)
def AltAz_to_ParticleFrame(altaz_, particleframe_):

    alt = altaz_.alt.rad
    az = altaz_.az.rad
    loc = altaz_.location
    time = altaz_.obstime
    r = u.Quantity(
        100, u.m
    )  # Warning dummy r value! For SphericalRepresentation to PhysicsSphericalRepresentation conversion

    conv_angle = Angle(convergence_angle(loc.lat.rad, loc.lon.rad), unit=u.radian)

    phi = np.pi / 2 + np.pi + conv_angle.rad - az

    theta = alt + np.pi / 2

    particleframe_ = ParticleFrame(
        phi=phi * rad, theta=theta * rad, obstime=time, location=loc, r=r
    )

    return particleframe_


# UTM: X = Easting, Y = Northing with  Northing = North + conv_angle
# Altaz: X = North Y = East


@frame_transform_graph.transform(FunctionTransform, UTM, AltAz)
def UTM_to_AltAz(UTM_, altaz):

    az = UTM_.azimuth.rad
    ze = UTM_.zenith.rad

    loc = UTM_.location
    time = UTM_.obstime

    conv_angle = Angle(convergence_angle(loc.lat.rad, loc.lon.rad), unit=u.radian)

    altitude = np.pi / 2 - ze
    Corrected_azimuth = (np.pi / 2 - az + conv_angle.rad) % (2 * np.pi)

    altaz = AltAz(
        alt=altitude * rad, az=Corrected_azimuth * rad, obstime=time, location=loc
    )

    return altaz


@frame_transform_graph.transform(FunctionTransform, AltAz, UTM)
def AltAz_to_UTM(altaz_, UTM_):

    alt = altaz_.alt.rad
    caz = altaz_.az.rad

    loc = altaz_.location
    time = altaz_.obstime
    r = u.Quantity(
        100, u.m
    )  # Warning dummy r value! For SphericalRepresentation to PhysicsSphericalRepresentation.

    conv_angle = Angle(convergence_angle(loc.lat.rad, loc.lon.rad), unit=u.radian)

    az = 5 * np.pi / 2 + conv_angle.rad - caz
    az = az % (2 * np.pi)
    ze = np.pi / 2 - alt

    UTM_ = UTM(azimuth=az * rad, zenith=ze * rad, obstime=time, location=loc, r=r)
    return UTM_


@frame_transform_graph.transform(FunctionTransform, UTM, ParticleFrame)
def UTM_to_ParticleFrame(UTM_, particleframe):

    phi = UTM_.azimuth.rad - np.pi
    theta = np.pi - UTM_.zenith.rad
    loc = UTM_.location
    time = UTM_.obstime
    r = UTM_.r

    particleframe = ParticleFrame(
        phi=phi * rad, theta=theta * rad, obstime=time, location=loc, r=r
    )

    return particleframe


@frame_transform_graph.transform(FunctionTransform, ParticleFrame, UTM)
def ParticleFrame_to_UTM(particleframe, UTM_):

    az = particleframe.phi.rad + np.pi
    ze = np.pi - particleframe.theta.rad

    loc = particleframe.location
    time = particleframe.obstime
    r = particleframe.r

    UTM_ = UTM(azimuth=az * rad, zenith=ze * rad, obstime=time, location=loc, r=r)

    return UTM_
